/*
 * Copyright 2022 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "kylin-tablet-desktop-backend.h"
#include "appnamesort.h"
#include <QString>
#include <KDesktopFile>
#include <KConfigGroup>
#include <QFileInfo>
#include <QEventLoop>
#include <QLocale>
#include <QCoreApplication>
#include <QTranslator>
#include <QDir>
#include <QDBusConnection>
#include <QDebug>
#include "ukuisettings.h"

const char *desktop_dir = "/usr/share/applications/";
// static int LeftIconSize = 19; //图标大小
const char *TABLET_LAUNCHER_SCHEMA = "org.kylin-tablet-desktop";
const char *TABLET_LAUNCHER_PATH = "/org/kylin-tablet-desktop/";

namespace UKUI {

kylinTabletDesktopBackend::kylinTabletDesktopBackend():
    m_idlePosPage(-2), m_appListener(nullptr),m_Filewatcher(nullptr)
  , m_maxPage(-1), m_loadCount(0), m_notEmptyDir(1)
  , m_panelSettings(nullptr), isFirstStart(false)
{
    qRegisterMetaType<appChangeInfo>("appChangeInfo");

    QTranslator *translator = new QTranslator(this);
    if(translator->load("/usr/share/kylin-tablet-desktop-backend/translations/" + QLocale::system().name())) {
        QCoreApplication::installTranslator(translator);
    }
    else
        qDebug() << "翻译文件加载失败！";

    // const QByteArray &schema_id, const QByteArray &path = QByteArray(), QObject *parent = NULL
    QByteArray id(TABLET_LAUNCHER_SCHEMA);
    if (QGSettings::isSchemaInstalled(id)) {
        m_tabletLauncher = new QGSettings(id, QByteArray(TABLET_LAUNCHER_PATH));
        m_line = m_tabletLauncher->get("line").toUInt();
        m_column = m_tabletLauncher->get("column").toUInt();
        isFirstStart = m_tabletLauncher->get("isFirstStart").toBool();
        if(isFirstStart) {
            qDebug() << "kylin tablet desktop is started firstly ! ! !";
            m_tabletLauncher->set("isFirstStart", QVariant(false));
        }
    }
    else {
        m_line = 4;
        m_column = 6;
    }

    QString dbPath = QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/ukui/appList.db";
    if ( !QFile::exists(dbPath) ) {
        qDebug() << "kylin tablet desktop is started firstly ! ! !";
        isFirstStart = true;    // 当手动将数据库删除时，也认定为是第一次启动
    }
    m_icon_theme = IconTheme::getInstance();
    loadPositions();
    desktopFileWatcher();
    panelConfFileWatcher();
    // 等待数据遍历完成后再继续初始化数据
    QEventLoop loop;
    connect(this, &kylinTabletDesktopBackend::firstTraversalCompleted, &loop, &QEventLoop::quit);
    loop.exec();
    // 获取数据库中的所有被删除应用的 id，要将这部分应用从数据库中删除
    // 防止在桌面崩溃后，用户通过其他手段将应用删除，此时检测是为了及时更新数据库
    checkIfDatabaseExistDeletedApp();
    monitorIconThemeChange();

    // 加载任务栏的应用数据，如果有变化的话，同步更改到数据库
    loadTaskBarData();
}

kylinTabletDesktopBackend::~kylinTabletDesktopBackend()
{
    if (m_baseItemList.size() > 0) {
        for (int i = 0; i < m_baseItemList.size(); i++) {
            delete m_baseItemList[i];
        }
    }
    if (m_appListener) {
        m_appListener->deleteLater();
        m_appListener = nullptr;
    }
    if (m_Filewatcher) {
        m_Filewatcher->deleteLater();
        m_Filewatcher = nullptr;
    }
    if (m_database) {
        m_database->deleteLater();
        m_database = nullptr;
    }
    if (m_icon_theme) {
        m_icon_theme->deleteLater();
        m_icon_theme = nullptr;
    }
    if (m_panelSettings) {
        m_panelSettings->deleteLater();
        m_panelSettings = nullptr;
    }
}

void kylinTabletDesktopBackend::loadPositions()
{
    m_database = new dataBaseOperation(this);
    m_database->openDataBase();
    m_baseItemList = m_database->queryAllItemData();

    allPage = m_database->queryAllPageData();
    allFlipSets = m_database->queryAllFlipSetData();
    allScrollSets = m_database->queryAllScrollSetData();
    for (int page = 0; page < allPage.size(); page++) {
        m_allItemList += allPage[allPage.keys().at(page)];
        if (page > m_maxPage)
            m_maxPage = page;
    }
    for (int i = 0; i < allFlipSets.size(); i++) {
        QList<QList<quint32>> appListList = allFlipSets[allFlipSets.keys().at(i)];
        for (int j = 0; j < appListList.size(); j++)
            m_allItemList += appListList[j];
    }
    for (int i = 0; i < allScrollSets.size(); i++) {
        m_allItemList += allScrollSets[allScrollSets.keys().at(i)];
    }
    qDebug() << "m_baseItemList =" << m_baseItemList;
    qDebug() << "allFlipSets =" << allFlipSets;
    qDebug() << "allScrollSets =" << allScrollSets;
    qDebug() << "m_allItemList = " << m_allItemList;
    qDebug() << "m_maxPage = " << m_maxPage;
    // 更新第一个空闲位置的 Item 所在的页数
    updateExistIdlePage();
}

void kylinTabletDesktopBackend::loadTaskBarData()
{
    // 获取任务栏的应用列表
    QList<quint32> taskBarAppData;
    if (isFirstStart)
    {
        if (QFile::exists(QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/ukui/panel.conf"))
        {
            m_panelSettings = new Settings(QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/ukui/panel.conf", QSettings::IniFormat, "taskbar");
            if (m_panelSettings) {
                QList<QMap<QString, QVariant> > taskBarApp = m_panelSettings->readArray("apps");
                for (int var = 0; var < taskBarApp.size(); ++var) {
                    quint32 taskbarAppId = getAppIdAccordingString(taskBarApp[var]["desktop"].toString());
                    if (taskbarAppId != 0) {
                        qDebug() << taskBarApp[var]["desktop"].toString() << taskbarAppId;
                        taskBarAppData.append(taskbarAppId);
                    }
                }
            }
        }
        else if (QFile::exists("/usr/share/ukui/panel.conf"))
        {
            m_panelSettings = new Settings("/usr/share/ukui/panel.conf", QSettings::IniFormat, "taskbar");
            if (m_panelSettings) {
                QList<QMap<QString, QVariant> > taskBarApp = m_panelSettings->readArray("apps");
                for (int var = 0; var < taskBarApp.size(); ++var) {
                    quint32 taskbarAppId = getAppIdAccordingString(taskBarApp[var]["desktop"].toString());
                    if (taskbarAppId != 0) {
                        qDebug() << taskBarApp[var]["desktop"].toString() << taskbarAppId;
                        taskBarAppData.append(taskbarAppId);
                    }
                }
            }
        }
        else
        {
            quint32 taskbarAppId;
            taskbarAppId = getAppIdAccordingString("/usr/share/applications/peony.desktop");
            if (taskbarAppId != 0)
                taskBarAppData.append(taskbarAppId);
            taskbarAppId = getAppIdAccordingString("/usr/share/applications/kylin-software-center.desktop");
            if (taskbarAppId != 0)
                taskBarAppData.append(taskbarAppId);
            taskbarAppId = getAppIdAccordingString("/usr/share/applications/wps-office-wps.desktop");
            if (taskbarAppId != 0)
                taskBarAppData.append(taskbarAppId);
        }

        if(m_database->querySetDataFromID(0).length() ==0 ){
            m_database->addScrollSet(0, taskBarAppData);
        }
        else{
            m_database->modifyScrollSetData(0, taskBarAppData);
        }

        allScrollSets[0] = taskBarAppData;
        qDebug() << "为了防止page页里有重复的id，需要将任务栏的应用从page页移除";
        // 为了防止page页里有重复的id，需要将任务栏的应用从page页移除
        for (int var = 0; var < allScrollSets[0].size(); ++var)
        {
            // 从 allPage 数据库删除该 item
            for (int pageId = 0; pageId < allPage.size(); pageId++)
            {
                int id = allScrollSets[0][var];
                QList<quint32> &itemlist = allPage[pageId];
                if (itemlist.indexOf(id) >= 0)
                {
                    itemlist.removeAt(itemlist.indexOf(id));
                    if (itemlist.size() == 0)
                        deleteLauncherPage(pageId);
                    else
                        modifyPageData(pageId, itemlist);
                    break;
                }
            }
        }
        qDebug() << "将任务栏的应用从page页移除后需要将前面page页的空白位置补齐";
        // 将任务栏的应用从page页移除后需要将前面page页的空白位置补齐
        updateExistIdlePage();
        for (quint32 pageId = m_idlePosPage; pageId < allPage.size();)
        {
            if (pageId != allPage.size() - 1) {
                quint32 nextPageId = pageId + 1;
                quint32 idleNum = updateExistIdlePage();
                QList<quint32> &currentPage = allPage[pageId];
                QList<quint32> &nextPage = allPage[nextPageId];
                quint32 i = 0;
                int j = 0; // 非应用的数量-因为可能存在分组
                // TODO: 当下一页仅剩一个分组，且当前页未填充满时，无法再从下一页移动应用填充满当前页。
                while(i < idleNum && j < nextPage.size())
                {
                    if (m_baseItemList[nextPage[j]]->getType() == Type::Icon)
                    {
                        currentPage.append(nextPage[j]);
                        nextPage.removeAt(j);
                        ++i;
                    }
                    else
                        ++j;
                }
                if (i != idleNum && j == 0) {
                    // 前面的页没有被补全，且下一页不存在分组。
                    // 说明下一页已经为空了，也没有把上一页补全。
                    // 删除下一页，并继续向前补全。
                    m_database->modifyPageData(pageId, currentPage);
                    deleteLauncherPage(nextPageId);
                    continue;
                }
                // 1. 前一页没有补全，但是下一页是分组，不再向前补齐；
                // 2. 前一页已经补全；
                m_database->modifyPageData(pageId, currentPage);
                if (nextPage.size() == 0)
                    deleteLauncherPage(nextPageId);
                else
                    m_database->modifyPageData(pageId+1, nextPage);
                ++pageId;
            }
            else {
                if (allPage[pageId].size() == 0)
                    deleteLauncherPage(pageId);
                else
                    m_database->modifyPageData(pageId, allPage[pageId]);
                ++pageId;
            }
        }
        for (int pageId = allPage.size() - 1; pageId >= 0; --pageId) {
            // 由于移动应用产生空白页，则删除
            if (allPage[pageId].size() == 0) {
                allPage.remove(pageId);
                m_database->deletePageData(pageId);
            }
        }
    }
    else if (QFile::exists(QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/ukui/panel.conf"))
    {
        m_panelSettings = new Settings(QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/ukui/panel.conf", QSettings::IniFormat, "taskbar");
        if (m_panelSettings) {
            QList<QMap<QString, QVariant> > taskBarApp = m_panelSettings->readArray("apps");
            for (int var = 0; var < taskBarApp.size(); ++var) {
                quint32 taskbarAppId = getAppIdAccordingString(taskBarApp[var]["desktop"].toString());
                qDebug() << taskBarApp[var]["desktop"].toString() << taskbarAppId;

                for (auto pageid = 0; pageid < allPage.count(); pageid++) {
                    for (auto itemid = 0; itemid < allPage[pageid].count(); itemid++) {
                        if (allPage[pageid].at(itemid) == taskbarAppId) {
                            allPage[pageid].removeAt(itemid);
                            m_database->modifyPageData(pageid, allPage[pageid]);
                        }
                    }
                }

                taskBarAppData.append(taskbarAppId);
            }
        }
        if (taskBarAppData != allScrollSets[0])
            taskBarAppChange(taskBarAppData);
    }
}

void kylinTabletDesktopBackend::monitorIconThemeChange()
{
    connect(m_icon_theme, &IconTheme::iconThemeChanged, [=](){
        for (int i = 0; i < m_baseItemList.size(); i++) {
            quint32 id = m_baseItemList.keys().at(i);
            if (Type::Icon == m_baseItemList[id]->getType())
            {
                IconItem *appItem = static_cast<IconItem *>(m_baseItemList[id]);
                QString absoluteDesktopName = appItem->getDesktopName();

                const KDesktopFile desktopFile(absoluteDesktopName);
                QString iconName = desktopFile.readIcon();

                QString absoluteIconName = getAbsoluteIconPath(iconName, 96);
                appItem->setIconName(absoluteIconName);
            }
        }
        emit iconThemeChanged();
    });
    emit iconThemeChanged();
}

void kylinTabletDesktopBackend::desktopFileWatcher()
{
    // 先获取到所有的应用列表，然后通过遍历所有系统中的desktop文件，存在系统中的应用则从此列表删除
    // 绝大多数情况，遍历完所有desktop文件后，删除列表应该为空
    // 但是也要防止在PC模式删除了一些应用，但是平板桌面插件还没有加载，这时候就会有一部分应用的删除监测不到
    m_shoudBeRemovedApp = m_baseItemList.keys();
    QList<quint32>::iterator pos = m_shoudBeRemovedApp.begin();
    while(pos != m_shoudBeRemovedApp.end()) {
        if(m_baseItemList[*pos]->getType() == Type::Icon)
        {
            pos++;
        }
        else
        {
            pos = m_shoudBeRemovedApp.erase(pos);
        }
    }
    m_appListener = new KDirLister(this);
    //    connect(m_appListener, &KCoreDirLister::started, this, &ListModelManager::onStarted);
    // 启动平板桌面后台时，会遍历一遍所有的desktop文件
    connect(m_appListener, &KCoreDirLister::itemsAdded, this, &kylinTabletDesktopBackend::onItemsAdded);
    connect(m_appListener, &KCoreDirLister::itemsDeleted, this, &kylinTabletDesktopBackend::onItemsDeleted);
    m_appListener->openUrl(QUrl::fromUserInput(QString(desktop_dir)), KDirLister::OpenUrlFlag::Keep);
    QDir android_path(QStandardPaths::standardLocations(QStandardPaths::ApplicationsLocation)[0]);
    if(android_path.exists()) {
        m_appListener->openUrl(QUrl::fromUserInput(QStandardPaths::standardLocations(QStandardPaths::ApplicationsLocation)[0]), KDirLister::OpenUrlFlag::Keep);
        if(!android_path.isEmpty())
        {
            m_notEmptyDir++;
        }
    }
    m_appListener->setAutoUpdate(true);
    m_appListener->started(QUrl::fromUserInput(QString(desktop_dir)));
}

void kylinTabletDesktopBackend::panelConfFileWatcher()
{
     m_Filewatcher = new QFileSystemWatcher(this);
     m_Filewatcher->addPath(QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/ukui/panel.conf");
     connect(m_Filewatcher, SIGNAL(fileChanged(const QString &)), this, SLOT(taskBarAppChange()));

}

void kylinTabletDesktopBackend::addLauncherPage(int page)
{
    m_maxPage = page;
    m_idlePosPage = m_maxPage;
    QList<quint32> itemList;
    allPage[m_idlePosPage] = itemList;
    m_database->addPage(m_idlePosPage, itemList);
//    emit pageAdded(m_idlePosPage);
}

void kylinTabletDesktopBackend::deleteLauncherPage(int page)
{
//    if(allPage[page].size() != 0)
//        return;
    qDebug() << "kylinTabletDesktopBackend::deleteLauncherPage(int page)" << page;
    for (quint32 j = page; j < m_maxPage; ++j) {
        // 将 page 页后面的 itemList 以此前移
        QList<quint32> item_list = allPage[j+1];
        modifyPageData(j, item_list);
    }
    // 删除最后一页，因为最后一页的内容已经拷贝到上一页
    allPage.remove(m_maxPage);
    m_database->deletePageData(m_maxPage);
    m_maxPage--;
    emit pageDeleted(page);
}

void kylinTabletDesktopBackend::removeLauncherItem(quint32 id)
{
    qDebug() << "kylinTabletDesktopBackend::removeLauncherItem(quint32 id)" << id;
    // 1. 从内存 m_allItemList 列表里删除该 id
    m_allItemList.removeOne(id);

    // 2. 从数据库 itemList 删除
    m_database->deleteItemListData(id);
    // warning，注意防止内存泄露
    delete m_baseItemList[id];
    if ( m_baseItemList.remove(id) != 1 )
        qDebug() << "移除 item" << "id" << "失败！";

    // 3. 检查 sets 数据库，是否存在删除的 item（不用考虑 set 全部为空的情况，但是FlipSet某一子页为空时，需要删除此子页）
    bool isDeleteIconFromSet = false;
    bool isDeleteIconFromPage = false;
    int key;
    for (int i = 0; i < allScrollSets.size(); i++)
    {
        key = allScrollSets.keys().at(i); // key值为set的id
        QList<quint32> all_scrollSetValue = allScrollSets[key];
        if(all_scrollSetValue.indexOf(id) >= 0)
        {
            all_scrollSetValue.removeAt(all_scrollSetValue.indexOf(id));
            modifyScrollSetData(key, all_scrollSetValue);
            qDebug() << "emit itemDeletedFromSet(id, key)" << id << key;
            emit itemDeletedFromSet(id, key);
            break;
        }
    }
    for (int i = 0; i < allFlipSets.size(); i++)
    {
        key = allFlipSets.keys().at(i); // key值为set的id
        QList<QList<quint32>> all_flipSetValue = allFlipSets[key];
        for(int j = 0; j < all_flipSetValue.size(); j++)
        {
            QList<quint32> &set_subpage_value = all_flipSetValue[j];
            if(set_subpage_value.indexOf(id) >= 0) {
                if(set_subpage_value.size() == 1) { // 说明该 set 子页存在值为id的item，删掉后此页为空
                    all_flipSetValue.removeAt(j);
                    modifyFlipSetData(key, all_flipSetValue);
                    isDeleteIconFromSet = true;
                    emit setsPageDeleted(key);
                    break;
                }
                else {
                    set_subpage_value.removeAt(set_subpage_value.indexOf(id));
                    modifyFlipSetData(key, all_flipSetValue);
                    qDebug() << "emit itemDeletedFromSet(id, key)" << id << key;
                    isDeleteIconFromSet = true;
                    emit itemDeletedFromSet(id, key);
                    break;
                }
            }
        }
        if(isDeleteIconFromSet) {
            break;
        }
    }

    // 4. 从 allPage 数据库删除该 item（需要考虑 page 为空, 然后删除空白 page 页的情况）
    for(int pageId = 0; pageId < allPage.size(); pageId++)
    {
        QList<quint32> &itemlist = allPage[pageId];
        if(itemlist.indexOf(id) >= 0)
        {
            itemlist.removeAt(itemlist.indexOf(id));
            if(itemlist.size() == 0) {
                emit itemDeletedFromPage(id, pageId);
//                deleteLauncherPage(pageId);
                isDeleteIconFromPage = true;
            }
            else {
                modifyPageData(pageId, itemlist);
                isDeleteIconFromPage = true;
                emit itemDeletedFromPage(id, pageId);
            }
            break;
        }
    }
#if 0
    // 0. 从 m_setApp 列表里删除该 id
    m_setsApp.removeOne(id);
    // 1. 从 itemList 数据库删除
    m_database->deleteItemListData(id);
    if( m_baseItemList.remove(id) == 1 )
        qDebug() << "移除 item" << "id" << "成功！";
    else
        qDebug() << "移除 item" << "id" << "失败！";
    bool isDeleteIcon = false;
    // 2. 检查 sets 数据库，是否存在删除的 item（不用考虑 page 页为空）
    for(int i = 0; i < allSets.size(); i++)
    {
        int key = allSets.keys().at(i);
        QList<QList<quint32>> all_pageValue = allSets[key];
        for(int j = 0; j < all_pageValue.size(); j++)
        {
            QList<quint32> &value = all_pageValue[j];
            if(value.indexOf(id) >= 0) {
                if(value.size() == 1) { // 说明该page页存在值为id的item，删掉后此页为空
                    all_pageValue.removeAt(j);
                    modifySetData(key, all_pageValue);
                    isDeleteIcon = true;
                    emit setsPageDeleted(key);
                    break;
                }
                else {
                    value.removeAt(value.indexOf(id));
                    modifySetData(key, all_pageValue);
                    isDeleteIcon = true;
                    emit itemDeletedFromSet(id, key);
                    qDebug() << "emit itemDeletedFromSet(id, key)" << id << key;
                    break;
                }
            }
        }
        if(isDeleteIcon) {
            qDebug() << "Delete Sets item: " << all_pageValue;
            break;
        }
    }
    // 3. 若 sets 没有该 item，则从 allPage 数据库删除该 item（需要考虑 page 为空, 然后删除空白 page 页的情况）
    if(!isDeleteIcon)
    {
        for(int i = 0; i < allPage.size(); i++)
        {
            int pageId = allPage.keys().at(i);
            QList<quint32> itemlist = allPage[pageId];
            if(itemlist.indexOf(id) >= 0)
            {
                itemlist.removeAt(itemlist.indexOf(id));
                if(itemlist.size() == 0) {
                    deleteLauncherPage(pageId);
                    // TODO: 不清楚是否有问题，先记录一下。
                    // 当 page 页只存在一个待删除的 item 后，只触发删除 page 的信号，不再触发删除 item 的信号
                    emit pageDeleted(pageId);
                }
                else {
                    modifyPageData(pageId, itemlist);
                    emit itemDeletedFromPage(id, pageId);
                    qDebug() << "emit itemDeletedFromPage(id, key)" << id << pageId;
                }
                break;
            }
        }
    }
#endif
}

quint32 kylinTabletDesktopBackend::getAppIdAccordingString(const QString desktopName)
{
    QMap<quint32, BaseItem *>::iterator i = m_baseItemList.begin();
    while (i != m_baseItemList.end()) {
        if (i.value()->getType() == Type::Icon) {
            IconItem *app = static_cast<IconItem *>(i.value());
            if (app->getDesktopName() == desktopName)
                return i.key();
        }
        ++i;
    }
    return 0;
}

QStringList kylinTabletDesktopBackend::getBlackListApp()
{
    QStringList blackList;
    blackList << "software-properties-livepatch.desktop"
              << "mate-color-select.desktop"
              << "blueman-adapters.desktop"
              << "mate-user-guide.desktop"
              << "nm-connection-editor.desktop"
              << "debian-uxterm.desktop"
              << "debian-xterm.desktop"
              << "im-config.desktop"
              << "fcitx.desktop"
              << "fcitx-configtool.desktop"
              << "onboard-settings.desktop"
              << "info.desktop"
              << "ukui-power-preferences.desktop"
              << "ukui-power-statistics.desktop"
              << "software-properties-drivers.desktop"
              << "software-properties-gtk.desktop"
              << "gnome-session-properties.desktop"
              << "org.gnome.font-viewer.desktop"
              << "xdiagnose.desktop"
              << "gnome-language-selector.desktop"
              << "mate-notification-properties.desktop"
              << "transmission-gtk.desktop"
              << "mpv.desktop"
              << "system-config-printer.desktop"
              << "org.gnome.DejaDup.desktop"
              << "yelp.desktop"
                 //v10
              << "mate-about.desktop"
              << "time.desktop"
              << "network.desktop"
              << "shares.desktop"
              << "mate-power-statistics.desktop"
              << "display-im6.desktop"
              << "display-im6.q16.desktop"
              << "openjdk-8-policytool.desktop"
              << "kylin-io-monitor.desktop"
              << "wps-office-uninstall.desktop"
              << "wps-office-misc.desktop"
              << "kylin-installer.desktop";
    return blackList;
}

bool kylinTabletDesktopBackend::matchingAppCategories(QString category, QStringList categorylist)
{
    int index;
    for (index = 0; index < categorylist.count(); index++) {
        if (category.contains(categorylist.at(index), Qt::CaseInsensitive)) {
            return true;
        }
    }
    if (index == categorylist.count()) {
        return false;
    }
    return false;
}

QList<quint32> kylinTabletDesktopBackend::sortAppIdAccordingLetter(QList<quint32> &unorderedIdList)
{
    QMap<AppNameSort, quint32> orderedApp;
    foreach (auto &id, unorderedIdList) {
        if(m_baseItemList[id]->getType() == Type::Icon && !m_baseItemList[id]->getName().isEmpty()) {
            QString appName = m_baseItemList[id]->getName();
            orderedApp[AppNameSort(appName)] = id;
        }
    }
    //    foreach (auto &app, orderedApp.keys()) {
    //        qDebug() << app.getAppName();
    //    }
    //    qDebug() << __func__ << unorderedIdList << orderedApp.values();
    return orderedApp.values();
}

void kylinTabletDesktopBackend::checkIfDatabaseExistDeletedApp()
{
    qDebug() << "需要删除的数据库应用为 m_shoudBeRemovedApp: " << m_shoudBeRemovedApp;
    foreach (quint32 id, m_shoudBeRemovedApp) {
        removeLauncherItem(id);
    }
}

QString kylinTabletDesktopBackend::getAbsoluteIconPath(QString iconName, quint32 size)
{
    QFileInfo iconFileInfo(iconName);
    QString absoluteIconPath;
    if (iconFileInfo.isFile() && (iconName.endsWith(".png") || iconName.endsWith(".svg") || iconName.endsWith(".ico"))) {
        return iconName;
    } else {
        iconName.remove(".png");
        iconName.remove(".svg");
        iconName.remove(".ico");
        absoluteIconPath = m_icon_theme->getPathFromName(iconName, size);

        if (absoluteIconPath.isEmpty()) {
            if (QFile::exists(QString("/usr/share/icons/hicolor/scalable/apps/%1.%2").arg(iconName).arg("svg"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/scalable/apps/%1.%2").arg(iconName).arg("svg");
            } else if (QFile::exists(QString("/usr/share/icons/hicolor/96x96/apps/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/96x96/apps/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/hicolor/64x64/apps/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/64x64/apps/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/hicolor/48x48/apps/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/48x48/apps/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/hicolor/16x16/apps/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/16x16/apps/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/hicolor/96x96/mimetypes/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/96x96/mimetypes/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/hicolor/64x64/mimetypes/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/64x64/mimetypes/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/hicolor/48x48/mimetypes/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/48x48/mimetypes/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/hicolor/16x16/mimetypes/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/hicolor/16x16/mimetypes/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/gnome/48x48/apps/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/gnome/48x48/apps/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/icons/gnome/32x32/apps/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/icons/gnome/32x32/apps/%1.%2").arg(iconName).arg("png");
            } else if (QFile::exists(QString("/usr/share/pixmaps/%1.%2").arg(iconName).arg("svg"))) {
                absoluteIconPath = QString("/usr/share/pixmaps/%1.%2").arg(iconName).arg("svg");
            } else if (QFile::exists(QString("/usr/share/pixmaps/%1.%2").arg(iconName).arg("png"))) {
                absoluteIconPath = QString("/usr/share/pixmaps/%1.%2").arg(iconName).arg("png");
            } else {
                absoluteIconPath = m_icon_theme->getPathFromName("application-x-desktop", size);
            }
        }
    }
    return absoluteIconPath;
}

QMap<Categories, QString> kylinTabletDesktopBackend::getAllAppCategories() {
    QMap<Categories, QString> categoriesInfo;
    categoriesInfo[Categories::Android] = tr("Mobile");
    categoriesInfo[Categories::Network] = tr("Internet");
    categoriesInfo[Categories::Messaging] = tr("Social");
    categoriesInfo[Categories::Audio] = tr("Video");
    categoriesInfo[Categories::Development] = tr("Development");
    categoriesInfo[Categories::Graphics] = tr("Image");
    categoriesInfo[Categories::Game] = tr("Game");
    categoriesInfo[Categories::Office] = tr("Office");
    categoriesInfo[Categories::Education] = tr("Education");
    categoriesInfo[Categories::System] = tr("System");
    categoriesInfo[Categories::Other] = tr("Others");
    return categoriesInfo;
}

QList<quint32> kylinTabletDesktopBackend::getAppIdListAccordingCategory(Categories type)
{
    QList<quint32> appIdList;
    QList<quint32> allItemId = m_baseItemList.keys();
    QList<BaseItem *> allItemInfo = m_baseItemList.values();
    for(int i = 0; i < m_baseItemList.size(); i++) {
        if(allItemInfo.at(i))
        {
            QList<int> categories = allItemInfo.at(i)->getCategories();
            if(categories.contains(static_cast<int32_t>(type)))
                appIdList.append(allItemId.at(i));
        }
    }
    return sortAppIdAccordingLetter(appIdList);
}

QList<quint32> kylinTabletDesktopBackend::getAllAppIdList()
{
    QList<quint32> appIdList;
    QList<quint32> allItemId = m_baseItemList.keys();
    for(int i = 0; i < m_baseItemList.size(); i++) {
        if( m_baseItemList[allItemId[i]] && (m_baseItemList[allItemId[i]]->getType() == Type::Icon) )
            appIdList.append(allItemId[i]);
    }
    return sortAppIdAccordingLetter(appIdList);
}

void kylinTabletDesktopBackend::onItemsAdded(const QUrl &directoryUrl, const KFileItemList &items)
{
    Q_UNUSED(directoryUrl)
    // copy from ukui-menu master
    QVector<QStringList> vector;
    vector.append(QStringList() << "Android");              // 0安卓
    vector.append(QStringList() << "Network");              // 1网络
    vector.append(QStringList() << "Messaging");            // 2社交
    vector.append(QStringList() << "Audio" << "Video");     // 3影音
    vector.append(QStringList() << "Development");          // 4开发
    vector.append(QStringList() << "Graphics");             // 5图像
    vector.append(QStringList() << "Game");                 // 6游戏
    vector.append(QStringList()
                  << "Office" << "Calculator"
                  << "Spreadsheet" << "Presentation"
                  << "WordProcessor" << "TextEditor");      // 7办公
    vector.append(QStringList() << "Education");            // 8教育
    vector.append(QStringList()
                  << "System"
                  << "Settings"
                  << "Security");                           // 9系统

    QStringList blackAppList = getBlackListApp();
    for (const KFileItem &fileItem : items) {
        if(!fileItem.isDesktopFile()) {
            continue;
        }
        if (!fileItem.url().toString().endsWith("desktop"))
        {
            continue;
        }

        QString desktopName = fileItem.name();
        QString absoluteDesktopName = fileItem.url().path();
        const KDesktopFile desktopFile(absoluteDesktopName);
        KConfigGroup desktopGroup =  desktopFile.desktopGroup();
        // 如果代码里没有写NoDisplay=false，desktopFile.noDisplay()返回的值也是true，所以不能用desktopFile.noDisplay()进行判断
        // 有些desktop写的是NoDisplay=true;多了一个分号，所以要用contains判断
        if(desktopGroup.readEntry("NoDisplay").contains("true") || desktopFile.readIcon().isEmpty() || !desktopFile.hasApplicationType()) {
            continue;
        }
        if(blackAppList.contains(desktopName))
            continue;

        QString iconName = desktopFile.readIcon();
        // appName 会随着系统语言的切换自动修改为当前使用的语言的名字
        QString appName = desktopFile.readName();
        QString appType = desktopFile.readType();
        QString execPath = desktopGroup.readEntry("Exec");
        QString app_categories = desktopGroup.readEntry("Categories");
        QString absoluteIconName = getAbsoluteIconPath(iconName, 96);

        bool is_owned = false;
        QList<int> actual_categories;
        for (int j = 0; j < vector.size(); j++) {
            if (matchingAppCategories(app_categories, vector.at(j))) { // 有对应分类
                is_owned = true;
                actual_categories.append(j);
            }
        }
        if (!is_owned) { // 该应用无对应分类
            actual_categories.append(vector.size());
        }

        //        qDebug() << app_categories << actual_categories << iconName << absoluteIconName;

        Q_UNUSED(appType)
        Q_UNUSED(execPath)
        //        qDebug() << endl << "desktopName =" << desktopName << "iconName =" << iconName << "appName =" << appName << "execPath =" << execPath;
        bool isFindIcon = false;
        for(int i = 0; i < m_baseItemList.size(); i++) {
            quint32 id = m_baseItemList.keys().at(i);
            if(m_baseItemList[id] && (Type::Icon == m_baseItemList[id]->getType())) {
                IconItem *appItem = static_cast<IconItem *>(m_baseItemList[id]);
                // qDebug() << __LINE__ << itemList->size() << appItem->getDesktopName() << desktopName;
                if(appItem->getDesktopName() == absoluteDesktopName) {
                    // 更新为当前语言的名字
                    if(appItem->getName() != appName || appItem->getCategories() != actual_categories || appItem->getIconName() != absoluteIconName)
                    {
                        appItem->setName(appName);
                        appItem->setCategories(actual_categories);
                        //                        qDebug() << "absoluteIconName: " << absoluteIconName;
                        appItem->setIconName(absoluteIconName);
                        m_database->modifyItemData(appItem->getId(), absoluteDesktopName, appName, absoluteIconName, actual_categories);
                    }

                    isFindIcon = true;
                    // 检测到系统存在该应用，将该应用从删除列表中移除
                    m_shoudBeRemovedApp.removeOne(appItem->getId());
                    break;
                }
            }
            if(isFindIcon)
                break;
        }

        if(isFindIcon)
            continue;
        int idleNumber = updateExistIdlePage();  // 更新第一个空闲位置的 Item 所在的页数

        qDebug() << __FUNCTION__ << "m_idlePosPage: " << m_idlePosPage
                 << "m_maxPage:" << m_maxPage
                 << "updateExistIdlePage()" << idleNumber;
        if(m_idlePosPage == -2) {   // 说明所有页已经填充满
            addLauncherPage(m_maxPage + 1);
        }
        // 添加到 item 列表数据库
        //        // 数据库存储的应用图标都是不加绝对路径的图标
        //        m_database->addItem(absoluteDesktopName, appName, iconName, static_cast<int32_t>(Type::Icon), {1,1}, actual_categories);
        // 将数据库存储的应用图标改为绝对路径
        m_database->addItem(getIdleItemID(), absoluteDesktopName, appName, absoluteIconName, static_cast<int32_t>(Type::Icon), {1,1}, actual_categories);
        quint32 id = m_database->queryIdFromDesktopName(absoluteDesktopName);
        // 如果该id已经存在page或者set里，就不再重复加入
        // 因为可能会在数据库提前插入一些预制的 id 进 page 或者 set
        if(!m_allItemList.contains(id)){
            // 添加到 page 列表数据库
            QList<quint32> &itemList = allPage[m_idlePosPage];
            itemList.append(id);
            m_database->modifyPageData(m_idlePosPage, itemList);
            m_allItemList.append(id);
        }

        BaseItem *baseItem = new UKUI::IconItem(absoluteDesktopName, appName, id, absoluteIconName, actual_categories);
        m_baseItemList[id] = baseItem;
        emit itemAdded(id, m_idlePosPage);
        qDebug() << "添加 item id 成功！";
        //        updateExistIdlePage();  // 更新第一个空闲位置的 Item 所在的页数
    }

    m_loadCount ++;
    if(m_loadCount == m_notEmptyDir) {
        emit firstTraversalCompleted();
    }
    //    if(isFirstLoad) {
    //        emit firstTraversalCompleted();
    //        isFirstLoad = false;
    //    }
}

void kylinTabletDesktopBackend::onItemsDeleted(const KFileItemList &items)
{
    qDebug() << __func__ << items.size();
    for (const KFileItem &fileItem : items)
    {
        if(!fileItem.isDesktopFile()) {
            continue;
        }
        if (!fileItem.url().toString().endsWith("desktop"))
        {
            continue;
        }
        //        QString desktopName = fileItem.name();
        QString absoluteDesktopName = fileItem.url().path();
        bool isFindIcon = false;
        quint32 id = 0;
        for(int i = 0; i < m_baseItemList.size(); i++) {
            id = m_baseItemList.keys().at(i);
            if(Type::Icon == m_baseItemList[id]->getType()) {
                IconItem *appItem = static_cast<IconItem *>(m_baseItemList[id]);
                if(appItem->getDesktopName() == absoluteDesktopName) {
                    isFindIcon = true;
                    break;
                }
            }
        }
        if(isFindIcon) {
            removeLauncherItem(id);
            updateExistIdlePage();
        }
    }
}

quint32 kylinTabletDesktopBackend::updateExistIdlePage()
{
    m_idlePosPage = -2;
    for(quint32 page = 0; page < allPage.size(); page++)
    {
        quint32 sum = 0;
        QList<quint32>::iterator pageItem = allPage[page].begin();
        while( pageItem != allPage[page].end())
        {
            quint32 itemId = *pageItem;
            BaseItem *item;
            if(m_baseItemList.find(itemId) != m_baseItemList.end())
            {
                item = m_baseItemList[itemId];
                pageItem ++;
            }
            else
            {
                m_database->updatePageItem();
                pageItem = allPage[page].erase(pageItem);
                continue;
            }

            sum += item->getPlaceHolder().at(0) * item->getPlaceHolder().at(1);
        }

        if(sum < m_line * m_column) {
            m_idlePosPage = page;
            qDebug() << "\tSum:" << sum << "Page:" << page << "Size:" << allPage[page].size() << "m_idlePosPage:" << m_idlePosPage;
            return m_line * m_column - sum;
        }
//        qDebug() << "\tSum:" << sum << "Page:" << page << "Size:" << allPage[page].size() << "m_idlePosPage:" << m_idlePosPage;
    }
    return 0;
}

void kylinTabletDesktopBackend::updateTaskBarAppList()
{

}

quint32 kylinTabletDesktopBackend::getIdleItemID()
{
    if (m_database) {
        return m_database->getIdleItemID();
    }
    else {
        return 0;
    }
}

void kylinTabletDesktopBackend::createGroup(quint32 groupId, QList<quint32> itemIdList, const QString groupName, bool scrollable)
{
    QString name;
    if (groupName.size() == 0) {
        QList<int> categoryList;
        qDebug() << " m_baseItemList number is : " << m_baseItemList.count() <<
                    "itemIdList[1] is :" << itemIdList[1];
        if (m_baseItemList.find(itemIdList[1]) != m_baseItemList.end())
        {
            categoryList = m_baseItemList[itemIdList[1]]->getCategories();
            QMap<Categories, QString> categoriesInfo = getAllAppCategories();
            name = categoriesInfo[Categories(categoryList.at(0))];
        }
        else
        {
            qDebug() << "m_baseItemList not find itemIdList[1] , Array out of bounds, use default name";
            name = "system";
        }
    }
    else
        name = groupName;
    //    // 新建 group 分组的 item
    //    quint32 groupId = m_database->getIdleItemID();
    m_database->addItem(groupId, "none", name, "none", 2).toUInt();
    m_baseItemList[groupId] = new UKUI::GroupItem(name, groupId);
    m_allItemList.append(groupId);
    if (scrollable) {
        m_database->addScrollSet(groupId, itemIdList);
        allScrollSets[groupId] = itemIdList;
    }
    else {
        QList<QList<quint32>> list;
        list.append(itemIdList);
        m_database->addFlipSet(groupId, list);
        allFlipSets[groupId] = list;
    }
    return;
}

void kylinTabletDesktopBackend::deleteGroup(quint32 setId)
{
    if(setId == 0)
    {
        return ;
    }
    m_database->deleteSetsData(setId);
    m_database->deleteItemListData(setId);
    m_database->updatePageItem();
}

QList<quint32> kylinTabletDesktopBackend::readPanelConf()
{
    QList<quint32> taskBarAppData;
    QString fileName(QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/ukui/panel.conf");
    // 如果通过vim或者编辑器修改无法监听到变化，所以必须将之前的QSettings析构掉，再重新new一个。
    if (m_panelSettings) {
        delete m_panelSettings;
        m_panelSettings = nullptr;
    }
    if (QFile::exists(fileName))
        m_panelSettings = new Settings(fileName, QSettings::IniFormat, "taskbar");

    if (m_panelSettings) {
        QList<QMap<QString, QVariant> > taskBarApp = m_panelSettings->readArray("apps");
        for (int var = 0; var < taskBarApp.size(); ++var) {
            quint32 taskbarAppId = getAppIdAccordingString(taskBarApp[var]["desktop"].toString());
            qDebug() << taskBarApp[var]["desktop"].toString() << taskbarAppId;
            taskBarAppData.append(taskbarAppId);
        }
        qDebug() << __func__ << __LINE__ << taskBarAppData;
    }
    return taskBarAppData;
}

void kylinTabletDesktopBackend::taskBarAppChange()
{
    m_Filewatcher->addPath(QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/ukui/panel.conf");
    qDebug() << "任务栏配置文件发生改变！" << __FUNCTION__;

    QList<quint32> taskBarAppData;
    taskBarAppData = readPanelConf();
    if( allScrollSets[0] == taskBarAppData )
        return ;

    appChangeInfo appChangePositionData;
    bool isExchange = true;
    qDebug() << "原本任务栏列表为：" << allScrollSets[0] <<"新的任务栏列表为:" << taskBarAppData;
    // 固定应用到任务栏
    for (int i = 0; i < taskBarAppData.size(); i++)
    {
        if ((! allScrollSets[0].contains(taskBarAppData.at(i))))
        {
            qDebug() << "固定应用到任务栏";
            isExchange = false;
            appChangePositionData.isDesktopDeleteApp = true;
            appChangePositionData.isTaskbarDeleteApp = false;
            appChangePositionData.isExchangePosition = false;
            appChangePositionData.itemId = taskBarAppData.at(i);
            //查询当前itemid 对应的页数
            for (int j = 0; j < allPage.size(); j++)
            {
                if (allPage[j].indexOf(taskBarAppData.at(i)) != -1)
                {
                    //计算出应用位于当前什么页的什么位置
                    appChangePositionData.itemFromPage = j;
                    appChangePositionData.itemFromPageIndex = allPage[j].indexOf(taskBarAppData.at(i));
                    appChangePositionData.itemToPage = 0;
                    appChangePositionData.itemToPageIndex = i;
                    break;
                }
            }

            appChangePositionData.itemToSet = 0;
            appChangePositionData.itemFromSet = -1;


            QList<quint32> &itemlist = allPage[appChangePositionData.itemFromPage];
            itemlist.removeAll(appChangePositionData.itemId);
            if(itemlist.size() == 0) {
                for (quint32 j = appChangePositionData.itemFromPage; j < m_maxPage; ++j) {
                    // 将 page 页后面的 itemList 依次前移
                    QList<quint32> item_list = allPage[j+1];
                    modifyPageData(j, item_list);
                }
                // 删除最后一页，因为最后一页的内容已经拷贝到上一页
                allPage.remove(m_maxPage);
                m_database->deletePageData(m_maxPage);
                m_maxPage--;
                updateExistIdlePage();
            }

            emit taskBarAppChanged(appChangePositionData);
        }
    }

    // 从任务栏取消应用的固定
    // warning: 后台发给前台的的位置移动信号是一个一个发送的，相应的前台也是一个一个处理移动信号的。
    //          所以会出现的一个问题是：
    //          如果任务栏应用列表为 7 9 4 2 5，2和5从任务栏取消固定，后台会先将2的index 3发送给前台，然后前台根据index删除数据
    //          然后将5的index 4发送给前台，因为前台已经删除了2，所以前台的最大index变为了3，所以后台发给前台的index 4会导致前台索引
    //          出错。
    //
    //          解决办法：
    //              后台发送一个信号后就处理一下索引信息，跟新一下后台内存中的应用列表，下个移动信号发送处理后的index
    for (int i = 0; i < allScrollSets[0].size();)
    {
        if (! taskBarAppData.contains(allScrollSets[0].at(i)))
        {
            qDebug() << "从任务栏取消应用固定";
            isExchange = false;
            appChangePositionData.isDesktopDeleteApp = false;
            appChangePositionData.isTaskbarDeleteApp = true;
            appChangePositionData.isExchangePosition = false;
            appChangePositionData.itemId = allScrollSets[0].at(i);
            //查询空白页的空白位置
            int pageNullPosition = updateExistIdlePage();
            if (pageNullPosition != 0)
            {
                appChangePositionData.itemToPage = m_idlePosPage;
                appChangePositionData.itemToPageIndex = 24-pageNullPosition;
            }
            else
            {
                qDebug() << "创建页之前，页数为" << allPage.size();
                addLauncherPage(m_maxPage + 1);
                qDebug() << "创建完页之后，页数为" << m_maxPage + 1;
                appChangePositionData.itemToPage = m_idlePosPage;
                appChangePositionData.itemToPageIndex = 0;
            }

            appChangePositionData.itemFromPage = 0;
            appChangePositionData.itemFromPageIndex = i;

            appChangePositionData.itemFromSet=0;
            appChangePositionData.itemToSet=-1;

            allPage[m_idlePosPage].append(appChangePositionData.itemId);
            m_database->modifyPageData(m_idlePosPage, allPage[m_idlePosPage]);
            allScrollSets[0].removeAll(appChangePositionData.itemId);

            emit taskBarAppChanged(appChangePositionData);
        }
        else
            ++i;
    }

    //任务栏交换应用
    if(isExchange)
    {
        qDebug() << "进行任务栏应用的交换";
        appChangePositionData.isDesktopDeleteApp = false;
        appChangePositionData.isTaskbarDeleteApp = false;
        appChangePositionData.isExchangePosition = true;
        appChangePositionData.itemFromPage = 0;
        appChangePositionData.itemToPage = 0;
        appChangePositionData.itemFromSet=0;
        appChangePositionData.itemToSet=0;
        int m=0;
        for(m = 0; m<taskBarAppData.size(); m++)
        {
            if(taskBarAppData.contains(allScrollSets[0].at(m)) && allScrollSets[0].indexOf(taskBarAppData[m]) != m)
                break;
        }

        bool isFrontMove = false;
        int i=m;

        while(i+1 <taskBarAppData.size() && allScrollSets[0].at(i) == taskBarAppData[i+1])
        {
            i++;
        }
        if(allScrollSets[0].at(i) ==  taskBarAppData[m])
        {
           isFrontMove = true;
           appChangePositionData.itemId = taskBarAppData[m];
           appChangePositionData.itemFromPageIndex = i;
           appChangePositionData.itemToPageIndex = m;
        }
        if (!isFrontMove)
        {
            while(i+1 <allScrollSets[0].size() && allScrollSets[0].at(i+1) == taskBarAppData[i])
            {
                i++;
            }
            if(allScrollSets[0].at(m) ==  taskBarAppData[i])
            {
               appChangePositionData.itemId = taskBarAppData[i];
               appChangePositionData.itemFromPageIndex = m;
               appChangePositionData.itemToPageIndex = i;
            }
        }

        emit taskBarAppChanged(appChangePositionData);
    }

    m_database->modifyScrollSetData(0, taskBarAppData);
    allScrollSets[0] = taskBarAppData;
}

void kylinTabletDesktopBackend::taskBarAppChange(QList<quint32> &taskBarAppData)
{
    qDebug() << "原本任务栏列表为：" << allScrollSets[0] <<"新的任务栏列表为:" << taskBarAppData;

    // 固定应用到任务栏
    for (int i = 0; i < taskBarAppData.size(); i++)
    {
        if ((! allScrollSets[0].contains(taskBarAppData.at(i))))
        {
            qDebug() << "固定应用到任务栏";
            //查询当前itemid 对应的页数
            for (int j = 0; j < allPage.size(); j++)
            {
                if (allPage[j].indexOf(taskBarAppData.at(i)) != -1)
                {
                    allPage[j].removeAll(taskBarAppData.at(i));
                    m_database->modifyPageData(j, allPage[j]);
                    if (0 == allPage[j].size())
                        deleteLauncherPage(j);
                    break;
                }
            }
        }
    }

    // 从任务栏取消应用的固定
    for (int i = 0;i < allScrollSets[0].size(); i++)
    {
        if (! taskBarAppData.contains(allScrollSets[0].at(i)))
        {
            qDebug() << "从任务栏取消应用固定";
            //查询空白页的空白位置
            int pageNullPosition = updateExistIdlePage();
            if (pageNullPosition == 0)
            {
                qDebug() << "创建页之前，页数为" << allPage.size();
                addLauncherPage(m_maxPage + 1);
                qDebug() << "创建完页之后，页数为" << m_maxPage + 1;
            }
            allPage[m_idlePosPage].append(allScrollSets[0].at(i));
            m_database->modifyPageData(m_idlePosPage, allPage[m_idlePosPage]);
        }
    }

    if (taskBarAppData != allScrollSets[0])
    {
        m_database->modifyScrollSetData(0, taskBarAppData);
        allScrollSets[0] = taskBarAppData;
    }
}

//quint32 kylinTabletDesktopBackend::createGroup(quint32 fromItemId, quint32 toItemId, const QString &groupName, bool scrollable)
//{
//    QList<quint32> m_applist;
//    m_applist.append(fromItemId);
//    m_applist.append(toItemId);
//    QString name;
//    if (groupName.size() == 0) {
//        QList<int> categoryList = m_baseItemList[toItemId]->getCategories();
//        QMap<Categories, QString> categoriesInfo = getAllAppCategories();
//        name = categoriesInfo[Categories(categoryList.at(0))];
//    }
//    else
//        name = groupName;
//    quint32 groupId = m_database->addItem(m_database->getIdleItemID(), "", name, "", 2).toUInt();
//    if (scrollable)
//        m_database->addScrollSet(groupId, m_applist);
//    else {
//        QList<QList<quint32>> list;
//        list.append(m_applist);
//        m_database->addFlipSet(groupId, list);
//    }
//    // TODO: true or false
//    return groupId;
//}

void kylinTabletDesktopBackend::modifyPageData(quint32 pageId, const QList<quint32> &itemIdList)
{
    if(itemIdList.size() == 0)
    {
        // 删除此 pageId 页，TODO
        allPage[pageId] = itemIdList;
        m_database->modifyPageData(pageId, itemIdList);
        deleteLauncherPage(pageId);
        return;
    }
    else {
        // 当传入的 id 页不存在时，先新建 page 页
        if (!allPage.keys().contains(pageId))
            addLauncherPage(pageId);
        allPage[pageId] = itemIdList;
        m_database->modifyPageData(pageId, itemIdList);
    }
}

void kylinTabletDesktopBackend::modifyFlipSetData(quint32 setId, const QList<QList<quint32>> &itemIdList)
{
    // TODO：
    // 当 sets 的类型为 group 时，内容为空时，应将 group 删除；container 时保留，不删除
    allFlipSets[setId] = itemIdList;
    m_database->modifyFlipSetData(setId, itemIdList);
}

void kylinTabletDesktopBackend::modifyScrollSetData(quint32 setId, const QList<quint32> &itemIdList)
{
    QList<quint32> itemSequence;
    for (int i=0; i < itemIdList.size(); i++)
    {
        if (m_database->queryItemIdFromList(itemIdList[i]))
        {
            itemSequence.append(itemIdList[i]);
        }
    }

    allScrollSets[setId] = itemSequence;
    m_database->modifyScrollSetData(setId, itemSequence);
    // 任务栏应用数据修改操作，写入 panel.conf 文件
    if (setId == 0)
    {
        QList<QMap<QString, QVariant> > hashList;

        for (int var = 0; var < itemSequence.size(); ++var) {
            QString desktopName;
            if (m_baseItemList.find(itemSequence[var]) != m_baseItemList.end())
            {
                desktopName = static_cast<IconItem *>(m_baseItemList[itemSequence[var]])->getDesktopName();
            }
            else
            {
                continue;
            }

            QHash<QString, QString> m_settingsMap;
            m_settingsMap["desktop"] = desktopName;
            // 下面三行是为了向group里添加file、exec、name的键值对，但默认为空；
            // 为了和任务栏数据保持一致，不加的话就不会添加该键值对；
            m_settingsMap["file"];
            m_settingsMap["exec"];
            m_settingsMap["name"];
            QMap<QString, QVariant> map;
            QHashIterator<QString, QString> it(m_settingsMap);
            while (it.hasNext())
            {
                it.next();
                map[it.key()] = it.value();
            }
            hashList << map;
        }
        //error
        if (m_panelSettings) {
            m_panelSettings->setArray("apps", hashList);
        }
    }
}

bool kylinTabletDesktopBackend::modifySetName(quint32 setId, QString name)
{
    if(name.isEmpty() || (m_baseItemList[setId]->getType() != Type::Container && m_baseItemList[setId]->getType() != Type::Group))
        return false;
    else {
        m_baseItemList[setId]->setName(name);
        return m_database->modifySetName(setId, name);
    }
}

}
