/*
 * Copyright 2022 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "modelmanager.h"
#include <QStack>
#include <QThread>
#include <QGuiApplication>
#include <QMetaEnum>
#include <QDebug>
#include "base/launcheritem.h"
#include "base/categoryitem.h"
#include "core/desktopdatahandler.h"
#include "core/appmanager.h"
#include "utils/gettaskbarvalue.h"

class ModelManagerPrivate : public QObject
{
    Q_DECLARE_PUBLIC(ModelManager)
public:
    ModelManagerPrivate(ModelManager *parent);
    enum ParentItemType {
        Page,
        Group,
    };
    struct PlaceholderItemInfo {
        LauncherItem *placeholderItem = nullptr;
        int itemIndex = -1;
        int pageIndex = -1;
        int groupId = -1;
        ParentItemType parentItemType;
        bool available() {
            return itemIndex > -1 &&
                    pageIndex > -1 &&
                    placeholderItem;
        }
    };

    ///
    /// \brief layoutForAddPlaceholderItem 如果添加 placeholder item 的页当前已经是满屏,
    ///  则需要将最后一个 item 移动到下一个屏幕,需要重新布局
    /// \param page 当前插入 placeholder item 的页
    /// \param groupId placeholder item 所在的分组 id,默认为 -1,即不在分组内
    ///
    void layoutForAddPlaceholderItem(int page, int groupId = -1);

    ///
    /// \brief layoutForRemovePlaceholderItem 删除 placeholder item 之后需要重新布局还原位置
    ///
    void layoutForRemovePlaceholderItem();

    ///
    /// \brief layoutForReplacePlaceholderItem 拖动 item 放至 placeholder item 之后,恢复由于插入
    ///  placeholder item 产生的布局变化
    ///
    void layoutForReplacePlaceholderItem();

    LauncherItem *createAddIconItem(int pageIndex, int state = LauncherItem::Normal);
    LauncherItem *createAppIcon(int itemId, int pageIndex);
    int itemIndex(int itemId, int page);
    PageModel *createPage(int pageIndex, int groupId, int state = LauncherItem::Editable);
    void tryAddAddIcon(int page, int groupId);

    QPair<QString, QString> categoryIcon(int category);
    bool itemExists(int groupId, int itemId);

    ///
    /// \brief itemGroupId 查找item所在的group的id
    /// \param itemId
    /// \return 如果item不在分组内则返回-1
    ///
    int itemGroupId(int itemId);
    QList<quint32> groupIds(int groupId) const;
    ///
    /// \brief removeGroupItem 主要用于添加界面的数据操作
    /// \param itemId
    /// \param groupId
    ///
    void removeGroupItem(int itemId, int groupId);

    void removeItemFromCategory(int itemId);
    int categoryOfAddtionPage(int page);

    QList<int> selectedIds() const;

    bool categoryContainsItem(int itemId, int category);

    ///
    /// \brief autoPushBackItemForGroup
    /// \param itemId
    /// \param groupId
    /// \param startPageIndex
    /// \return
    ///
    int autoPushBackItemForGroup(int itemId, int groupId, int startPageIndex);

    ///
    /// \brief updateCategoriesItemsPageIndex update the pageIndex of item in \memberof categoryItemsModel
    /// \param startPageIndex
    /// \param diff
    ///
    void updateCategoriesItemsPageIndex(int startPageIndex, int diff);

    void printCategoryItemPageIndex() {
        for (int page=0; page<categoryItemsModel.count(); ++page) {
            for (int itemIndex=0; itemIndex < categoryItemsModel.at(page)->count(); ++itemIndex) {
                auto item = categoryItemsModel.at(page)->at(itemIndex);
                qDebug() << item->name() << item->pageIndex();
            }
        }
    }

    void initSelectedItemIds(int groupId);

    int availableId();
//    PageModel* pagemodel;
//    MultiPageModel *groupdModel[0] = new MultiPageModel();

    MultiPageModel launcherModel;
    QMap<int, MultiPageModel *> groupdModel;

    // 保存每个分组的每一页可容纳的item的最大数量
    QMap<int, int> groupPageMaxItemCount;
    PlaceholderItemInfo placeholderItemInfo;

    CategoryModel categoryModel;
    MultiPageModel categoryItemsModel;

    //
    int currentCategoryOfAddion;

    QList<int> currentSelectedIds;
    const int kTaskBarGroupId = 0;

    /// \todo 根据配置文件修改配置
    int pageItemMaxCount = 24;
    bool needAddIcon = false;
    const int pageItemMaxOfAddition = 28;
    QList<int> editableContainerId = { 2, 4 ,0};
    ///

    // 由于 placeholder item 导致页面布局改变的页
    // 不包括当前放置 placeholder item 的页
    QStack<int> layoutChangedPages;

    DesktopDataHandler *dataHandler;
    AppManager *appManager;
    GetTaskbarValue *taskBarManager;

    ModelManager *q_ptr;
};

ModelManagerPrivate::ModelManagerPrivate(ModelManager *parent)
    : dataHandler(new DesktopDataHandler())
    , appManager(new AppManager())
    , taskBarManager(new GetTaskbarValue(parent))
    , q_ptr(parent)
{
    Q_Q(ModelManager);

    QThread *backendThread = new QThread;
    dataHandler->moveToThread(backendThread);
    connect(backendThread, &QThread::started, dataHandler, &DesktopDataHandler::init);
    connect(dataHandler, &DesktopDataHandler::finished, backendThread, &QThread::quit);
    connect(backendThread, &QThread::finished, dataHandler, &DesktopDataHandler::deleteLater);
    connect(backendThread, &QThread::finished, backendThread, &QThread::deleteLater);
    connect(qApp, &QGuiApplication::aboutToQuit, dataHandler, &DesktopDataHandler::finished);

    connect(dataHandler, &DesktopDataHandler::desktopData, q, &ModelManager::desktopData);
    connect(dataHandler, &DesktopDataHandler::categoriesData, q, &ModelManager::categories);

    connect(dataHandler, &DesktopDataHandler::appUninstalledFromGroup, q, &ModelManager::removeItemFromGroup);
    connect(dataHandler, &DesktopDataHandler::appUninstalled, q, &ModelManager::removeItemWithoutGroup);
    connect(dataHandler, &DesktopDataHandler::appInstalled, q, &ModelManager::itemInstalled);
#if 0
    connect(q, &ModelManager::itemMoved, dataBackend, &DataBackend::moveItem);
    connect(q, &ModelManager::itemRemoved, dataBackend, &DataBackend::removeItem);
    connect(q, &ModelManager::itemInserted, dataBackend, &DataBackend::insertItem);
#endif
    connect(q, &ModelManager::groupNameChanged, dataHandler, &DesktopDataHandler::setGroupName);
    connect(q, &ModelManager::updateGroup, dataHandler, &DesktopDataHandler::updateGroup);

    connect(dataHandler, &DesktopDataHandler::iconChanged, q, &ModelManager::iconThemeChanged);

    connect(dataHandler, &DesktopDataHandler::pageRemoved, q, &ModelManager::pageRemoved);

    connect(dataHandler, &DesktopDataHandler::taskbarDataChange, q, &ModelManager::changeTaskbarData);

    backendThread->start();
}

void ModelManagerPrivate::layoutForAddPlaceholderItem(int page, int groupId)
{
    layoutChangedPages.clear();
    MultiPageModel *model = groupId > 0 ? groupdModel.value(groupId) : &launcherModel;
    int maxPageCount = groupId > 0 ? groupPageMaxItemCount.value(groupId) : pageItemMaxCount;
    // 超出当前页的item的数量的时候，将最后一个item移动至下一页
    // 如果最后一个item的是添加按钮的话，直接删除即可
    if (model->at(page)->count() > maxPageCount) {
        int startPage = page;
        while (model->at(startPage)->count() > maxPageCount &&
               startPage<model->count()-1) {
            qDebug() << "start page " << startPage;
            auto moveToNextPageItem = model->at(startPage)->takeLast();
            if (moveToNextPageItem->type() == LauncherItem::AddIcon) {
                moveToNextPageItem->deleteLater();
                continue;
            }
            moveToNextPageItem->setPageIndex(moveToNextPageItem->pageIndex() + 1);

            model->at(++startPage)->prepend(moveToNextPageItem);
            layoutChangedPages.push(startPage);
        }
    }
    qDebug() << "layoutChangedPages" << layoutChangedPages;
}

void ModelManagerPrivate::layoutForRemovePlaceholderItem()
{
    LauncherItem *moveItem = nullptr;
    MultiPageModel *model = placeholderItemInfo.groupId > 0 ?
                groupdModel.value(placeholderItemInfo.groupId) : &launcherModel;
    while (!layoutChangedPages.isEmpty()) {
        int page = layoutChangedPages.pop();
        moveItem = model->at(page)->takeFirst();
        moveItem->setPageIndex(moveItem->pageIndex() - 1);
        model->at(--page)->append(moveItem);
    }
    // fix: 不显示“添加“按钮的分组内的item拖动至新建页时导致数组越界
#if 0
    if (model->last()->isEmpty()) {
        model->removeLast();
    }
#endif
}

void ModelManagerPrivate::layoutForReplacePlaceholderItem()
{
    LauncherItem *moveItem = nullptr;
    MultiPageModel *model = placeholderItemInfo.groupId > 0 ?
                groupdModel.value(placeholderItemInfo.groupId) : &launcherModel;
    //
    while (layoutChangedPages.count() > 1) {
        int page = layoutChangedPages.pop();
        moveItem = model->at(page)->takeFirst();
        moveItem->setPageIndex(moveItem->pageIndex() - 1);

        model->at(--page)->append(moveItem);
    }
    // fix: 不显示“添加“按钮的分组内的item拖动至新建页时导致数组越界
#if 0
    if (model->last()->isEmpty()) {
        model->removeLast();
    }
#endif
}

LauncherItem *ModelManagerPrivate::createAddIconItem(int pageIndex, int state)
{
    if (!needAddIcon) {
        return nullptr;
    }
    LauncherItem *addItem = new LauncherItem;
    addItem->setId(0);
    addItem->setPageIndex(pageIndex);
    addItem->setType(LauncherItem::AddIcon);
    addItem->setIcon("qrc:/ui/images/ukui_add_default.svg");
    addItem->setName("添加");
    addItem->setRowSpan(1);
    addItem->setColumnSpan(1);
    addItem->setState(state);
    return addItem;
}

LauncherItem *ModelManagerPrivate::createAppIcon(int itemId, int pageIndex)
{
    LauncherItem *item = new LauncherItem;
    QMap<quint32, UKUI::BaseItem *> items = dataHandler->items();
    UKUI::IconItem *iconItem = static_cast<UKUI::IconItem *>(items.value(itemId));
    if (iconItem) {
        item->setId(itemId);
        item->setPageIndex(pageIndex);
        item->setName(iconItem->getName());
        item->setIcon(iconItem->getIconName());
        item->setDesktopFile(iconItem->getDesktopName());
        item->setColumnSpan(1);
        item->setRowSpan(1);
    }

    return item;
}

int ModelManagerPrivate::itemIndex(int itemId, int page)
{
    Q_ASSERT(page >= 0);
    if (launcherModel.count() <= page) {
        return -1;
    }
    for (int index=0; index<launcherModel.at(page)->count(); ++index) {
        if (launcherModel.at(page)->at(index)->id() == itemId) {
            return index;
        }
    }
    return -1;
}

PageModel *ModelManagerPrivate::createPage(int pageIndex, int groupId, int state)
{
    PageModel *page = new PageModel;
    if (editableContainerId.contains(groupId)) {
        LauncherItem *addItem = createAddIconItem(pageIndex, state);
        page->push_back(addItem);
    }
    return page;
}

void ModelManagerPrivate::tryAddAddIcon(int page, int groupId)
{
    if (!needAddIcon) {
        return;
    }
    MultiPageModel *model = groupId > 0 ? groupdModel.value(groupId) : &launcherModel;
    if (model->at(page)->isEmpty()) {
        return;
    }

    int maxItemCount = groupId > 0 ? groupPageMaxItemCount.value(groupId) : pageItemMaxCount;
    if (model->at(page)->count() < maxItemCount &&
            model->at(page)->last()->type() != LauncherItem::AddIcon &&
            editableContainerId.contains(groupId)) {
        model->at(page)->push_back(createAddIconItem(page));
    }
}

QPair<QString, QString> ModelManagerPrivate::categoryIcon(int category)
{
    QMetaEnum metaEnum = QMetaEnum::fromType<CategoryItem::Category>();
    const char *categoryName = metaEnum.valueToKey(category);

    return QPair<QString, QString>(QString("qrc:") + "/ui/images/categories/ukui_" + categoryName + "_default.svg",
                                   QString("qrc:") + "/ui/images/categories/ukui_" + categoryName + "_click.svg");
}

bool ModelManagerPrivate::itemExists(int groupId, int itemId)
{
    Q_ASSERT(groupdModel.contains(groupId));
    int pageCount = groupdModel.value(groupId)->count();
    for (int page=0; page<pageCount; ++page) {
        for (int itemIndex=0;
             itemIndex<groupdModel.value(groupId)->at(page)->count();
             ++itemIndex) {
            if (groupdModel.value(groupId)->at(page)->at(itemIndex)->id() == itemId) {
                return true;
            }
        }
    }
    return false;
}

int ModelManagerPrivate::itemGroupId(int itemId)
{
    auto groupdModelIt = groupdModel.constBegin();
    while (groupdModelIt != groupdModel.constEnd()) {
        auto groupdModel = groupdModelIt.value();
        for (int page=0; page<groupdModel->count(); ++page) {
            for (int itemIndex=0; itemIndex<groupdModel->at(page)->count(); ++itemIndex) {
                if (groupdModel->at(page)->at(itemIndex)->id() == itemId) {
                    return groupdModelIt.key();
                }
            }
        }
        ++ groupdModelIt;
    }
    return -1;
}

QList<quint32> ModelManagerPrivate::groupIds(int groupId) const
{
    Q_ASSERT(groupdModel.contains(groupId));
    QList<quint32> ids;
    for (int page=0; page<groupdModel.value(groupId)->count(); ++page) {
        int pageItemCount = groupdModel.value(groupId)->at(page)->count();
        for (int index=0; index<pageItemCount; ++index) {
            ids.push_back(groupdModel.value(groupId)->at(page)->at(index)->id());
        }
    }
    return ids;
}

void ModelManagerPrivate::removeGroupItem(int itemId, int groupId)
{
    Q_ASSERT(groupdModel.contains(groupId));
    int pageCount = groupdModel.value(groupId)->count();
    for (int page=0; page<pageCount; ++page) {
        int pageItemCount = groupdModel.value(groupId)->at(page)->count();
        for (int itemIndex=0; itemIndex<pageItemCount; ++itemIndex) {
            if (groupdModel.value(groupId)->at(page)->at(itemIndex)->id() == itemId) {
                groupdModel.value(groupId)->at(page)->removeAt(itemIndex);
                tryAddAddIcon(page, groupId);
                return;
            }
        }
    }
}

void ModelManagerPrivate::removeItemFromCategory(int itemId)
{
    Q_Q(ModelManager);
    // 需要从全部和具体的分类中都删除
    int categoryItemPageCount = categoryItemsModel.count();
    for (int page=0; page<categoryItemPageCount; ++page) {
        int itemCountInPage = categoryItemsModel.at(page)->count();
        for (int itemIndex=0; itemIndex < itemCountInPage; ++itemIndex) {
            if (categoryItemsModel.at(page)->at(itemIndex)->id() == itemId) {
                categoryItemsModel.at(page)->removeAt(itemIndex);
                if (categoryItemsModel.at(page)->isEmpty()) {
                    int category = categoryOfAddtionPage(page);
                    updateCategoriesItemsPageIndex(page, -1);
                    categoryItemsModel.removeAt(page);
                    Q_EMIT q->categoryPageRemoved(category);
                    -- categoryItemPageCount;
                    // 如果当前分类下的应用全部被删除了，则删除该分类
                    for (int categoryIndex=0; categoryIndex<categoryModel.count(); ++categoryIndex) {
                        if (categoryModel.at(categoryIndex)->category() == category) {
                            if (categoryModel.at(categoryIndex)->pageCount() == 1) {
                                categoryModel.removeAt(categoryIndex);
                                Q_EMIT q->categoryRemoved(category);
                            } else if (categoryModel.at(categoryIndex)->pageCount() > 1) {
                                int newPageCount = categoryModel.at(categoryIndex)->pageCount() - 1;
                                categoryModel.at(categoryIndex)->setPageCount(newPageCount);
                            }
                        }
                    }
#if 0
                    int pageCount = 0;
                    for (int categoryIndex=0; categoryIndex<categoryModel.count(); ++categoryIndex) {
                        pageCount += categoryModel.at(categoryIndex)->pageCount();
                        if (pageCount > page) {
                            int newPageCount = categoryModel.at(categoryIndex)->pageCount() - 1;
                            if (newPageCount > 0) {
                                categoryModel.at(categoryIndex)->setPageCount(newPageCount);
                            } else {
                                int category = categoryModel.at(categoryIndex)->category();
                                categoryModel.removeAt(categoryIndex);
                                Q_EMIT q->categoryRemoved(category);
                            }
                            break;
                        }
                    }
#endif
                    break;
                }
                // move
                else {
                    while (page + 1 < categoryItemsModel.count() &&
                            categoryOfAddtionPage(page) == categoryOfAddtionPage(page + 1) &&
                            categoryOfAddtionPage(page) != -1 &&
                            !categoryItemsModel.at(page + 1)->isEmpty()) {
                        auto movedItem = categoryItemsModel.at(page + 1)->takeFirst();
                        movedItem->setPageIndex(movedItem->pageIndex() - 1);
                        categoryItemsModel.at(page ++)->push_back(movedItem);
                        if (categoryItemsModel.at(page)->isEmpty()) {
                            int category = categoryOfAddtionPage(page);
                            categoryItemsModel.removeAt(page);
                            for (int categoryIndex=0; categoryIndex<categoryModel.count(); ++categoryIndex) {
                                if (categoryModel.at(categoryIndex)->category() == category) {
                                    int newPageCount = categoryModel.at(categoryIndex)->pageCount() - 1;
                                    categoryModel.at(categoryIndex)->setPageCount(newPageCount);
                                }
                            }
                            Q_EMIT q->categoryPageRemoved(category);
                            -- categoryItemPageCount;
                            -- page;
                        }
                    }
                    break;
                }
            }
        }
    }
}

int ModelManagerPrivate::categoryOfAddtionPage(int page)
{
    Q_ASSERT(categoryItemsModel.count() > page && page >= 0);
    int pageStartIndex = 0, pageEndIndex = 0;
    for (int categoryIndex=0; categoryIndex<categoryModel.count(); ++categoryIndex) {
        int curCategoryCount = categoryModel.at(categoryIndex)->pageCount();
        pageStartIndex = pageEndIndex;
        pageEndIndex += curCategoryCount - 1;
        if (page >= pageStartIndex && page <= pageEndIndex) {
            return categoryModel.at(categoryIndex)->category();
        }
        ++ pageEndIndex;
    }
    return -1;
}

QList<int> ModelManagerPrivate::selectedIds() const
{
    QList<int> ids;
    for (int page=0; page<categoryItemsModel.count(); ++page) {
        int itemCount = categoryItemsModel.at(page)->count();
        for (int itemIndex=0; itemIndex<itemCount; ++itemIndex) {
            auto item = categoryItemsModel.at(page)->at(itemIndex);
            if (item->state() == LauncherItem::Selectable && item->selected()) {
                if (!ids.contains(item->id())) {
                    ids.push_back(item->id());
                }
            }
        }
    }
    return ids;
}

bool ModelManagerPrivate::categoryContainsItem(int itemId, int category)
{
    int startPage = 0, endPage = 0;
    bool foundCaegory = false;
    for (int categoryIndex=0; categoryIndex<categoryModel.count(); ++categoryIndex) {
        if (categoryModel.at(categoryIndex)->category() == category) {
            endPage = startPage + categoryModel.at(categoryIndex)->pageCount() - 1;
            foundCaegory = true;
            break;
        }
        startPage += categoryModel.at(categoryIndex)->pageCount();
    }
    if (foundCaegory) {
        for (int page=startPage; page<=endPage; ++page) {
            int itemCount = categoryItemsModel.at(page)->count();
            for (int index=0; index<itemCount; ++index) {
                if (categoryItemsModel.at(page)->at(index)->id() == itemId) {
                    return true;
                }
            }
        }
    }
    return false;
}

int ModelManagerPrivate::autoPushBackItemForGroup(int itemId, int groupId, int startPageIndex)
{
    Q_ASSERT(groupdModel.contains(groupId));
    if (groupdModel[groupId]->count() <= startPageIndex) {
        groupdModel[groupId]->push_back(createPage(startPageIndex, groupId, LauncherItem::Normal));
    }

    if (groupdModel[groupId]->at(startPageIndex)->count() < groupPageMaxItemCount[groupId]) {
        int curItemCount = groupdModel[groupId]->at(startPageIndex)->count();
        // -1,当前有添加按钮
        groupdModel[groupId]->at(startPageIndex)->insert(curItemCount-1, createAppIcon(itemId, startPageIndex));
        return startPageIndex;
    } else if (groupdModel[groupId]->at(startPageIndex)->count() == groupPageMaxItemCount[groupId]) {
        // 替换添加按钮
        if (groupdModel[groupId]->at(startPageIndex)->last()->type() == LauncherItem::AddIcon) {
            int index = groupdModel[groupId]->at(startPageIndex)->count() - 1;
            groupdModel[groupId]->at(startPageIndex)->replace(index, createAppIcon(itemId, startPageIndex));
            return startPageIndex;
        } else {
            autoPushBackItemForGroup(itemId, groupId, ++startPageIndex);
        }
    } else {
        autoPushBackItemForGroup(itemId, groupId, ++startPageIndex);
    }
    return startPageIndex;
}

void ModelManagerPrivate::updateCategoriesItemsPageIndex(int startPageIndex, int diff)
{
    Q_ASSERT(startPageIndex >=0 && categoryItemsModel.count() > startPageIndex &&
             diff > - categoryItemsModel.count() + 1);
    for (int page=startPageIndex; page<categoryItemsModel.count(); ++page) {
        for (int itemIndex=0; itemIndex < categoryItemsModel.at(page)->count(); ++itemIndex) {
            auto item = categoryItemsModel.at(page)->at(itemIndex);
            item->setPageIndex(item->pageIndex() + diff);
        }
    }
}

void ModelManagerPrivate::initSelectedItemIds(int groupId)
{
    Q_ASSERT(groupdModel.contains(groupId));
    currentSelectedIds.clear();
    for (int page=0; page<groupdModel.value(groupId)->count(); ++page) {
        int itemCount = groupdModel.value(groupId)->at(page)->count();
        for (int itemIndex=0; itemIndex<itemCount; ++itemIndex) {
            currentSelectedIds.push_back(groupdModel.value(groupId)->at(page)->at(itemIndex)->id());
        }
    }
}

int ModelManagerPrivate::availableId()
{
    QMap<int, int> ids;

    auto getIdFunction = [&](const MultiPageModel &model) {
        int launchPageCount = model.count();
        for (int page=0; page<launchPageCount; ++page) {
            int pageItemCount = model.at(page)->count();
            for (int itemIndex=0; itemIndex<pageItemCount; ++itemIndex) {
                int id = model.at(page)->at(itemIndex)->id();
                ids.insert(id, id);
            }
        }
    };

    getIdFunction(launcherModel);

    auto it = groupdModel.constBegin();
    while (it != groupdModel.constEnd()) {
        getIdFunction(**it);
        ++ it;
    }

    if (ids.lastKey() == ids.count() - 1) {
        return ids.count();
    }

    int idIndex = 0;
    auto idIt = ids.constBegin();
    while (idIt != ids.constEnd()) {
        if (idIndex != *idIt && idIndex > 0) {
            return idIndex;
        }
        ++ idIt;
        ++ idIndex;
    }

    return ids.count();
}


ModelManager::ModelManager(QObject *parent)
    : QObject(parent)
    , d_ptr(new ModelManagerPrivate(this))
{
    qRegisterMetaType<MultiPageModel>("MultiPageModel");
}

ModelManager::~ModelManager()
{
    delete d_ptr;
}

QAbstractListModel *ModelManager::launcherModel() const
{
    Q_D(const ModelManager);
    return const_cast<MultiPageModel *>(&d->launcherModel);
}

QAbstractListModel *ModelManager::groupModel(const quint32 &id) const
{
    Q_D(const ModelManager);
    return d->groupdModel[id];
}

QAbstractListModel *ModelManager::categoryModel() const
{
    Q_D(const ModelManager);
    return const_cast<CategoryModel *>(&d->categoryModel);
}

QAbstractListModel *ModelManager::categoryItemsModel() const
{
    Q_D(const ModelManager);
    return const_cast<MultiPageModel *>(&d->categoryItemsModel);
}

int ModelManager::groupPageItemCount(int groupId, int pageIndex)
{
    Q_D(ModelManager);

    MultiPageModel *model = groupId >= 0 ? d->groupdModel.value(groupId) : &d->launcherModel;
    return model->isEmpty() ? 0 : model->at(pageIndex)->count();
}

int ModelManager::launcerPageItemCount(int pageIndex) const
{
    Q_D(const ModelManager);
    Q_ASSERT(d->launcherModel.count() > pageIndex && pageIndex >= 0);
    return d->launcherModel.at(pageIndex)->count();
}

void ModelManager::moveItem(int from, int to, int page, int groupId)
{
    Q_D(ModelManager);
    if (from == to) {
        return;
    }

    MultiPageModel *model = groupId >= 0 ? d->groupdModel.value(groupId) : &d->launcherModel;
    model->at(page)->move(from, to);
}

void ModelManager::removeItem(int index, int page, int groupId)
{
    Q_D(ModelManager);
    MultiPageModel *model = groupId >= 0 ? d->groupdModel.value(groupId) : &d->launcherModel;
    int pageItemMaxCount = groupId >= 0 ? d->groupPageMaxItemCount.value(groupId) : d->pageItemMaxCount;
    model->at(page)->removeAt(index);

    if (groupId == 0) {
        updateCachedData(groupId);
        return;
    }
    // 当前页为空时删除该页
    if ((model->at(page)->isEmpty() && groupId != d->kTaskBarGroupId) ||
            (d->needAddIcon
             && model->at(page)->count() == 1
             && model->at(page)->last()->type() == LauncherItem::AddIcon)) {
        model->removeAt(page);
        for (int pageIndex=page; pageIndex<model->count(); ++pageIndex) {
            PageModel *pageModel = model->at(pageIndex);
            for (int itemIndex=0; itemIndex<pageModel->count(); ++itemIndex) {
                pageModel->at(itemIndex)->setPageIndex(pageModel->at(itemIndex)->pageIndex() - 1);
            }
        }
    } else if (d->needAddIcon) {
        if (model->at(page)->count() < pageItemMaxCount &&
                model->at(page)->last()->type() != LauncherItem::AddIcon &&
                d->editableContainerId.contains(groupId)) {
            LauncherItem *addItem = d->createAddIconItem(page, LauncherItem::Editable);
            model->at(page)->push_back(addItem);
        }
    }
    updateCachedData(groupId);
}

void ModelManager::setItemState(int state, int groupId)
{
    Q_D(ModelManager);
    Q_ASSERT(groupId >= -1);
    MultiPageModel *model = groupId >= 0 ? d->groupdModel.value(groupId) : &d->launcherModel;
    for (int page=0; page < model->count(); ++page) {
        auto pageModel = model->at(page);
        for (int itemIndex=0; itemIndex < pageModel->count(); ++itemIndex) {
            if(groupId == 0 && pageModel->at(itemIndex)->appOpen()) {
                continue;
            }
            pageModel->at(itemIndex)->setState(state);
        }
    }

    static bool hasCreateNewPage = false;
    switch (state) {
    case LauncherItem::Normal: {
        hasCreateNewPage = false;
        for (int page=0; page < model->count(); ++page) {
            if (model->count() > 1) {
                if ((model->at(page)->count() == 1
                     && model->at(page)->last()->type() == LauncherItem::AddIcon) ||
                        (model->at(page)->isEmpty())) {
                    model->removeAt(page);
                    // 更新当前页之后的所有item的pageIndex
                    for (int changePage=page; changePage<model->count(); ++changePage) {
                        for (int itemIndex=0; itemIndex<model->at(changePage)->count(); ++itemIndex) {
                            auto item = model->at(changePage)->at(itemIndex);
                            item->setPageIndex(item->pageIndex() - 1);
                        }
                    }
                    -- page;
                }
#ifdef EDU_SCREEN
                d->tryAddAddIcon(page, groupId);
#endif
            }
        }
#if 0
        if (!d->needAddIcon) {
            if (model->last()->isEmpty()) {
                model->removeLast();
            }
        } else {
            if ((model->last()->count() == 1
                    && model->last()->first()->type() == LauncherItem::AddIcon) ||
                    model->last()->isEmpty()) {
                model->removeLast();
            }
        }
#endif
        break;
    }
    case LauncherItem::Editable: {
        if (!hasCreateNewPage && groupId != 0) {
            model->push_back(d->createPage(model->count(), groupId));
            hasCreateNewPage = true;
        }
        break;
    }
    default:
        break;
    }
}

void ModelManager::setGroupName(const QString &name, int groupId)
{
    Q_D(ModelManager);
    Q_ASSERT(groupId > 0);

    for (int pageIndex=0; pageIndex < d->launcherModel.count(); ++pageIndex) {
        for (int itemIndex=0; itemIndex < d->launcherModel.at(pageIndex)->count(); ++itemIndex) {
            if ((d->launcherModel.at(pageIndex)->at(itemIndex)->type() == LauncherItem::Container
                 ||d->launcherModel.at(pageIndex)->at(itemIndex)->type() == LauncherItem::Group) &&
                    d->launcherModel.at(pageIndex)->at(itemIndex)->id() == groupId) {
                d->launcherModel.at(pageIndex)->at(itemIndex)->setName(name);
            }
        }
    }

    Q_EMIT groupNameChanged(name, groupId);
}

void ModelManager::setAddAppCategory(int category, int groupId)
{
    Q_D(ModelManager);
    Q_ASSERT(d->groupdModel.contains(groupId));
    d->currentCategoryOfAddion = category;
    d->initSelectedItemIds(groupId);

    int startPage = 0, endPage = 0;
    if (category == CategoryItem::All) {
        startPage = 0;
        endPage = d->categoryItemsModel.count() - 1;//d->categoryModel.first()->pageCount() - 1;
    } else {
        for (int index=0; index<d->categoryModel.count(); ++index) {
            if (d->categoryModel.at(index)->category() == category) {
                -- startPage;
                endPage = startPage + d->categoryModel.at(index)->pageCount();
                break;
            }
            startPage += d->categoryModel.at(index)->pageCount();
        }
    }
    for (int page=0; page<d->categoryItemsModel.count(); ++page) {
        int pageItemCount = d->categoryItemsModel.at(page)->count();
        for (int itemIndex=0; itemIndex<pageItemCount; ++itemIndex) {
            auto item = d->categoryItemsModel.at(page)->at(itemIndex);
            if (((category == CategoryItem::All
                  && !d->categoryContainsItem(item->id(), CategoryItem::Education)
                  && d->categoryOfAddtionPage(page) != CategoryItem::Education)
                 || (category != CategoryItem::All && d->dataHandler->categoriesItems()
                     .value(UKUI::Categories::Education).contains(item->id()))) &&
                    page >= startPage && page <= endPage) {
                item->setState(LauncherItem::Selectable);
                int itemGroupId = d->itemGroupId(item->id());
                if (itemGroupId == -1) {
                    item->setSelected(false);
                } else {
                    if (itemGroupId == groupId) {
                        item->setSelected(true);
                    } else {
                        item->setState(LauncherItem::Normal);
                    }
                }
            } else {
                item->setState(LauncherItem::Normal);
            }
        }
    }
}

void ModelManager::updateCachedData(int groupId)
{
    Q_D(ModelManager);
    Q_ASSERT(groupId >= -1);
    MultiPageModel *model = groupId >= 0 ? d->groupdModel.value(groupId) : &d->launcherModel;
    QList<QList<quint32>> groupItemIds;
    for (int page=0; page<model->count(); ++page) {
        QList<quint32> itemIds;
        for (int itemIndex=0;
             itemIndex < model->at(page)->count();
             ++ itemIndex) {
            int id = model->at(page)->at(itemIndex)->id();
#ifdef EDU_SCREEN
            // add app icon
            if (id == 0) {
                continue;
            }
#endif
            if (groupId == d->kTaskBarGroupId) {
                if (findDesktopName(model->at(page)->at(itemIndex)->desktopFile()).isEmpty()) {
                    itemIds.push_back(id);
                }
            } else {
                itemIds.push_back(id);
            }
        }
        if (!itemIds.isEmpty()) {
            groupItemIds.push_back(itemIds);
        }
    }
    Q_EMIT updateGroup(groupItemIds, groupId);
}

bool ModelManager::groupEditable(int groupId)
{
    Q_D(ModelManager);
    return d->editableContainerId.contains(groupId);
}

void ModelManager::updateGroupItems(int groupId, int pageIndex)
{
    Q_D(ModelManager);
    Q_ASSERT(d->groupdModel.contains(groupId));

    auto selectedIds = d->selectedIds();

    // add
    for (const auto &id : selectedIds) {
        if (!d->itemExists(groupId, id)) {
            pageIndex = d->autoPushBackItemForGroup(id, groupId, pageIndex);
            qDebug() << "insert page index: " << pageIndex;
        }
    }
    // remove
    auto groupIds = d->groupIds(groupId);
    for (const auto &id : groupIds) {
        // 添加按钮的id是0
        if (!selectedIds.contains(id) && id != 0) {
            d->removeGroupItem(id, groupId);
        }
    }

#if 0
    if (d->groupdModel.value(groupId)->last()->count() == d->groupPageMaxItemCount.value(groupId)) {
        d->groupdModel[groupId]->push_back(
                    d->createPage(d->groupdModel.value(groupId)->count(), groupId, LauncherItem::Normal));
    }
#endif

    updateCachedData(groupId);
}

void ModelManager::setItemSelectState(int itemId, int pageIndex, bool selected)
{
    Q_D(ModelManager);
    if (d->categoryItemsModel.isEmpty()) {
        return;
    }
    if (d->categoryItemsModel.count() <= pageIndex || pageIndex < 0) {
        return;
    }

    if (selected) {
        d->currentSelectedIds.push_front(itemId);
    } else {
        d->currentSelectedIds.removeOne(itemId);
    }

    for (int page=0; page<d->categoryItemsModel.count(); ++page) {
        int itemCount = d->categoryItemsModel.at(page)->count();
        for (int itemIndex=0; itemIndex<itemCount; ++itemIndex) {
            auto item = d->categoryItemsModel.at(page)->at(itemIndex);
            if (item->id() == itemId && item->state() == LauncherItem::Selectable) {
                item->setSelected(selected);
            }
        }
    }
}

void ModelManager::addPlaceholderItem(int index,
                                      int pageIndex,
                                      int rowSpan,
                                      int columnSpan,
                                      int groupId)
{
    Q_D(ModelManager);

    qDebug() << "Add placeholder item " << index << pageIndex << groupId;
    if (index < 0 || pageIndex < 0) {
        return;
    }
    // main page
    if (index >= d->pageItemMaxCount && groupId == -1) {
        -- index;
    }

    MultiPageModel *model = groupId >= 0 ? d->groupdModel.value(groupId) : &d->launcherModel;
    if (d->placeholderItemInfo.available()) {
        // 如果是在同一页的占位 item,则直接移动占位 item 即可
        if (d->placeholderItemInfo.pageIndex == pageIndex &&
                d->placeholderItemInfo.groupId == groupId) {
            moveItem(d->placeholderItemInfo.itemIndex, index, pageIndex, groupId);
            d->placeholderItemInfo.itemIndex = index;
            return;
        }
        // in the main page
        if (d->placeholderItemInfo.groupId == -1) {
            d->launcherModel.at(d->placeholderItemInfo.pageIndex)->removeAt(d->placeholderItemInfo.itemIndex);
        } else {
            model->at(d->placeholderItemInfo.pageIndex)->removeAt(d->placeholderItemInfo.itemIndex);
        }
    }
    LauncherItem *placeholderItem = new LauncherItem;
    d->placeholderItemInfo.placeholderItem = placeholderItem;
    d->placeholderItemInfo.parentItemType = groupId >= 0 ?
                ModelManagerPrivate::Group : ModelManagerPrivate::Page;
    d->placeholderItemInfo.itemIndex = index;
    d->placeholderItemInfo.pageIndex = pageIndex;
    d->placeholderItemInfo.groupId = groupId;
    qDebug() << "d->placeholderItemInfo.itemIndex1: " << d->placeholderItemInfo.itemIndex;

    placeholderItem->setRowSpan(rowSpan);
    placeholderItem->setColumnSpan(columnSpan);
    placeholderItem->setPageIndex(pageIndex);
    placeholderItem->setType(LauncherItem::Placeholder);

#ifdef EDU_SCREEN
    // 不可编辑的分组不存在添加按钮，所有需要push_back
    if (d->editableContainerId.contains(groupId)) {
        model->at(pageIndex)->insert(index, placeholderItem);
    } else {
        model->at(pageIndex)->push_back(placeholderItem);
        d->placeholderItemInfo.itemIndex = model->at(pageIndex)->count() - 1;
        qDebug() << "d->placeholderItemInfo.itemIndex2: " << d->placeholderItemInfo.itemIndex;
    }
#else
    model->at(pageIndex)->insert(index, placeholderItem);
#endif
    d->layoutForAddPlaceholderItem(pageIndex, groupId);
}

void ModelManager::removePlaceholderItem()
{
    Q_D(ModelManager);
    if (!d->placeholderItemInfo.available()) {
        return;
    }
    if (d->placeholderItemInfo.parentItemType == ModelManagerPrivate::Page) {
        d->launcherModel[d->placeholderItemInfo.pageIndex]->removeAt(d->placeholderItemInfo.itemIndex);
    } else if (d->placeholderItemInfo.parentItemType == ModelManagerPrivate::Group) {
        d->groupdModel[d->placeholderItemInfo.groupId]->
                at(d->placeholderItemInfo.pageIndex)->removeAt(d->placeholderItemInfo.itemIndex);
    }
    d->placeholderItemInfo.itemIndex = -1;
    d->placeholderItemInfo.pageIndex = -1;
    d->placeholderItemInfo.placeholderItem = nullptr;

    d->layoutForRemovePlaceholderItem();
}

void ModelManager::replacePlaceholderItem(int fromIndex, int fromPage, int groupId)
{
    Q_D(ModelManager);
    Q_ASSERT(fromIndex >=0 && fromPage >=0 && groupId >= -1);
    if (!d->placeholderItemInfo.available()) {
        return;
    }

    LauncherItem *moveItem = nullptr;
    MultiPageModel *model = groupId >= 0 ?
                d->groupdModel.value(groupId) : &d->launcherModel;

    moveItem = model->at(fromPage)->takeAt(fromIndex);
    moveItem->setPageIndex(d->placeholderItemInfo.pageIndex);
    model->at(d->placeholderItemInfo.pageIndex)->replace(
                d->placeholderItemInfo.itemIndex, moveItem);

#ifdef EDU_SCREEN
    d->tryAddAddIcon(fromPage, groupId);
    d->layoutForReplacePlaceholderItem();
#endif
    d->placeholderItemInfo.itemIndex = -1;
    d->placeholderItemInfo.pageIndex = -1;
    d->placeholderItemInfo.groupId = -1;
    d->layoutChangedPages.clear();
    d->placeholderItemInfo.placeholderItem->deleteLater();
    d->placeholderItemInfo.placeholderItem = nullptr;

#if 0
    if (model->at(fromPage)->isEmpty()) {
        model->removeAt(fromPage);
        for (int changePage=fromPage; changePage<model->count(); ++changePage) {
            for (int itemIndex=0; itemIndex<model->at(changePage)->count(); ++itemIndex) {
                auto item = model->at(changePage)->at(itemIndex);
                item->setPageIndex(item->pageIndex()-1);
            }
        }
    }
#endif
}
void ModelManager::replacePlaceholderItemTaskbar(int fromIndex, int fromPage, int fromgroupId, int togroupId)
{
    Q_D(ModelManager);
    if (!d->placeholderItemInfo.available()) {
        return;
    }
    if (d->placeholderItemInfo.groupId == fromgroupId) {
        return;
    }
    LauncherItem *moveItem = nullptr;
    MultiPageModel *model = fromgroupId >= 0 ?
                d->groupdModel.value(fromgroupId) : &d->launcherModel;

    moveItem = model->at(fromPage)->takeAt(fromIndex);
    moveItem->setPageIndex(d->placeholderItemInfo.pageIndex);

    if(togroupId == 0) {
        for(int itemIndex = 0; itemIndex < d->groupdModel[0]->at(0)->count(); itemIndex++) {
            if(d->groupdModel[0]->at(0)->at(itemIndex)->desktopFile() == moveItem->desktopFile()) {
                moveItem->setAppOpen(true);
                moveItem->setState(LauncherItem::Normal);
                moveItem->setWindowId(d->groupdModel[0]->at(0)->at(itemIndex)->windowId());
                if(d->placeholderItemInfo.itemIndex + 1 >= d->groupdModel[0]->at(0)->count()) {
                    d->placeholderItemInfo.itemIndex -= 1;
                }
                removeItem(itemIndex, 0, 0);
            }
        }
    }

    MultiPageModel *desktopModel = togroupId >= 0 ?
                d->groupdModel.value(togroupId) : &d->launcherModel;

    desktopModel->at(d->placeholderItemInfo.pageIndex)->replace(
                d->placeholderItemInfo.itemIndex, moveItem);

    d->placeholderItemInfo.itemIndex = -1;
    d->placeholderItemInfo.pageIndex = -1;
    d->placeholderItemInfo.groupId = -1;
    d->layoutChangedPages.clear();
    d->placeholderItemInfo.placeholderItem->deleteLater();
    d->placeholderItemInfo.placeholderItem = nullptr;
}

QAbstractListModel* ModelManager::taskbarModel() //const
{
    Q_D(ModelManager);
    if(!d->groupdModel.contains(0)) {
        MultiPageModel *multiPage = new MultiPageModel;
        multiPage->push_back(new PageModel);
        d->groupdModel[0] = multiPage;
    }
    return d->groupdModel[0]->at(0);
}

void ModelManager::setAppOpenState(QString desktopName, int wid)
{
    Q_D(const ModelManager);
    for(int itemIndex = 0; itemIndex < d->groupdModel[0]->at(0)->count(); itemIndex++) {
        if(d->groupdModel[0]->at(0)->at(itemIndex)->desktopFile().endsWith(desktopName)) {
            d->groupdModel[0]->at(0)->at(itemIndex)->setAppOpen(true);
            d->groupdModel[0]->at(0)->at(itemIndex)->setWindowId(wid);
            return;
        }
    }
    qDebug() << "desktopName cannot be found";
}

void ModelManager::setAppCloseState(QString desktopName)
{
    Q_D(const ModelManager);
    for(int itemIndex = 0; itemIndex < d->groupdModel[0]->at(0)->count(); itemIndex++) {
        if(d->groupdModel[0]->at(0)->at(itemIndex)->desktopFile().endsWith(desktopName)) {
            d->groupdModel[0]->at(0)->at(itemIndex)->setAppOpen(false);
            return;
        }
    }
    qDebug() << "desktopName cannot be found";
}

void ModelManager::addWindowIconToTaskbar(QString iconPath, QString desktopName, int wid)
{
    Q_D(ModelManager);
    LauncherItem* newTestItem = new LauncherItem();
    newTestItem->setId(d->groupdModel.value(0)->at(0)->count() + 100);
    newTestItem->setPageIndex(0);
    newTestItem->setColumnSpan(1);
    newTestItem->setRowSpan(1);
    newTestItem->setEditable(false);
    newTestItem->setType(LauncherItem::AppIcon);
    newTestItem->setIcon(iconPath);
    newTestItem->setDesktopFile(desktopName);
    newTestItem->setAppOpen(true);
    newTestItem->setWindowId(wid);
    d->groupdModel[0]->at(0)->append(newTestItem);
}

void ModelManager::removeWindowIconFromTaskbar(QString iconPath)
{
    Q_D(const ModelManager);
    for(int itemIndex = 0; itemIndex < d->groupdModel[0]->at(0)->count(); itemIndex++) {
        if(d->groupdModel[0]->at(0)->at(itemIndex)->icon() == iconPath) {
            int index = itemIndex;
            removeItem(index, 0, 0);
            return;
        }
    }
}

void ModelManager::removeAppIconFromTaskBar(const QString &desktopFileName)
{
    Q_D(const ModelManager);
    for(int itemIndex = 0; itemIndex < d->groupdModel[0]->at(0)->count(); itemIndex++) {
        if(d->groupdModel[0]->at(0)->at(itemIndex)->desktopFile().contains(desktopFileName)) {
            removeItem(itemIndex, 0, 0);
            return;
        }
    }
}

QString ModelManager::findIconPath(QString desktopName)
{
    Q_D(const ModelManager);
    for (int pageIndex=0; pageIndex < d->launcherModel.count(); ++pageIndex) {
        for (int itemIndex=0; itemIndex < d->launcherModel.at(pageIndex)->count(); ++itemIndex) {
            if ((d->launcherModel.at(pageIndex)->at(itemIndex)->desktopFile() == desktopName)) {
                 return d->launcherModel.at(pageIndex)->at(itemIndex)->icon();
            }
        }
    }
    return QString();
}

QString ModelManager::findDesktopName(QString desktopName)
{
    Q_D(const ModelManager);
    for (int pageIndex=0; pageIndex < d->launcherModel.count(); ++pageIndex) {
        for (int itemIndex=0; itemIndex < d->launcherModel.at(pageIndex)->count(); ++itemIndex) {
            if ((d->launcherModel.at(pageIndex)->at(itemIndex)->desktopFile().endsWith(desktopName))) {
//                qDebug() << "desktopName contains desktopname"<< d->launcherModel.at(pageIndex)->at(itemIndex)->desktopFile();
                return d->launcherModel.at(pageIndex)->at(itemIndex)->desktopFile();
            }
        }
    }
    return QString();
}
QString ModelManager::findDesktopNameFromTaskbar(QString desktopName)
{
    Q_D(const ModelManager);
    if (d->groupdModel.isEmpty()) {
        return QString();
    }
    for(int itemIndex = 0; itemIndex < d->groupdModel[0]->at(0)->count(); itemIndex++) {
        if(d->groupdModel[0]->at(0)->at(itemIndex)->desktopFile().contains(desktopName)) {
            return d->groupdModel[0]->at(0)->at(itemIndex)->desktopFile();
        }
    }
    return QString();
}

void ModelManager::activeWindow(quint32 wid)
{
    Q_D(const ModelManager);
    d->taskBarManager->appWindowOpen(wid);
}

void ModelManager::changeAppWindowWid(QString desktopName, int wid, int newWid)
{
    Q_D(const ModelManager);
    if (!findDesktopName(desktopName).isEmpty()) {
        for(int itemIndex = 0; itemIndex < d->groupdModel[0]->at(0)->count(); itemIndex++) {
            if (d->groupdModel[0]->at(0)->at(itemIndex)->desktopFile().contains(desktopName)) {
               if (d->groupdModel[0]->at(0)->at(itemIndex)->windowId() == wid) {
                   d->groupdModel[0]->at(0)->at(itemIndex)->setWindowId(newWid);
               }
            }
        }
    }
    else {
        for(int itemIndex = 0; itemIndex < d->groupdModel[0]->at(0)->count(); itemIndex++) {
            if(d->groupdModel[0]->at(0)->at(itemIndex)->desktopFile().contains(desktopName)) {
                if (d->groupdModel[0]->at(0)->at(itemIndex)->windowId() == wid) {
                    d->groupdModel[0]->at(0)->at(itemIndex)->setWindowId(newWid);
                }
            }
        }
    }
}

void ModelManager::setPageModel(const MultiPageModel &pageModel)
{
    Q_D(ModelManager);
    for (int i=0; i<pageModel.count(); ++i) {
        d->launcherModel.append(pageModel.at(i));
    }
}

void ModelManager::setGroupModel(const QMap<int, MultiPageModel *> &groupModel)
{
    Q_D(ModelManager);
    d->groupdModel = groupModel;
}

void ModelManager::desktopData(const QMap<quint32, QList<quint32> > &pageData,
                               const QMap<quint32, QList<QList<quint32>>> &setData,
                               const QMap<quint32, UKUI::BaseItem *> &itemData)
{
    Q_D(ModelManager);

    // taskbar
    if (setData.contains(0) && !setData.value(0).isEmpty()) {
        PageModel *groupPageModel = new PageModel;
        QList<quint32> childItemsId = setData.value(0).at(0);
        for (int index=0; index<childItemsId.size(); ++index) {
            LauncherItem *launcherItem = new LauncherItem;
            UKUI::IconItem *iconItem = static_cast<UKUI::IconItem *>(itemData.value(childItemsId.at(index)));
            if (!iconItem) {
                break;
            }
            launcherItem->setId(iconItem->getId());
            launcherItem->setName(iconItem->getName());
            launcherItem->setType(LauncherItem::AppIcon);
            launcherItem->setPageIndex(0);
            launcherItem->setIcon(iconItem->getIconName());
            launcherItem->setDesktopFile(iconItem->getDesktopName());
            launcherItem->setRowSpan(1);
            launcherItem->setColumnSpan(1);
            groupPageModel->push_back(launcherItem);
        }
        if (d->groupdModel.contains(0)) {
            for (auto item : *groupPageModel) {
                d->groupdModel[0]->at(0)->push_back(item);
            }
            groupPageModel->deleteLater();
        } else {
            MultiPageModel *groupModel = new MultiPageModel;
            groupModel->push_back(groupPageModel);
            d->groupdModel.insert(0, groupModel);
        }        
    }

    QMap<quint32, QList<quint32>>::const_iterator i = pageData.constBegin();
    while (i != pageData.constEnd()) {
        PageModel *pageModel = new PageModel;
        PageModel groupPageModel, containerPageModel;
        for (const auto &itemId : i.value()) {
            UKUI::BaseItem *item = itemData.value(itemId);
            LauncherItem *launcherItem = new LauncherItem;
            if (!launcherItem || !item) {
                continue;
            }
            switch (item->getType()) {
            case UKUI::Type::Icon: {
                UKUI::IconItem *iconItem = static_cast<UKUI::IconItem *>(item);
                if (!iconItem) {
                    break;
                }
                launcherItem->setId(iconItem->getId());
                launcherItem->setName(iconItem->getName());
                launcherItem->setType(LauncherItem::AppIcon);
                launcherItem->setPageIndex(i.key());
                launcherItem->setIcon(iconItem->getIconName());
                launcherItem->setDesktopFile(iconItem->getDesktopName());
                launcherItem->setRowSpan(1);
                launcherItem->setColumnSpan(1);
                break;
            }
            case UKUI::Type::Group: {
                UKUI::GroupItem *groupItem = static_cast<UKUI::GroupItem *>(item);
                if (!groupItem) {
                    break;
                }
                launcherItem->setId(groupItem->getId());
                launcherItem->setName(groupItem->getName());
                launcherItem->setType(LauncherItem::Group);
                launcherItem->setPageIndex(i.key() - 1);
                launcherItem->setRowSpan(1);
                launcherItem->setColumnSpan(1);
                PageModel *groupPageModel = new PageModel;
                MultiPageModel *groupModel = new MultiPageModel;
                QList<QList<quint32>> childItemsId = setData.value(groupItem->getId());
                groupModel->reserve(childItemsId.size() / 9 + 1);
                for (int index=0; index<childItemsId.size(); ++index) {
                    if (index % 9 == 0) {
                        groupModel->push_back(groupPageModel);
                        groupPageModel = new PageModel;
                    }
                    LauncherItem *groupItem = new LauncherItem;
                    groupPageModel->push_back(groupItem);
                }
                if (groupItem->getId() == 0 && d->groupdModel.contains(0)) {
                    for (auto item : *groupPageModel) {
                        d->groupdModel[0]->at(0)->push_back(item);
                    }
                    groupModel->deleteLater();
                    groupPageModel->deleteLater();
                } else {
                    groupModel->push_back(groupPageModel);
                    d->groupdModel.insert(groupItem->getId(), groupModel);
                }

                break;
            }

            case UKUI::Type::Container: {
                UKUI::ContainerItem *containerItem = static_cast<UKUI::ContainerItem *>(item);
                if (!containerItem) {
                    break;
                }
                launcherItem->setId(containerItem->getId());
                launcherItem->setName(containerItem->getName());
                launcherItem->setType(LauncherItem::Container);
                launcherItem->setPageIndex(i.key() - 1);
                launcherItem->setRowSpan(containerItem->getPlaceHolder().at(0));
                launcherItem->setColumnSpan(containerItem->getPlaceHolder().at(1));

                /// \todo 根据配置设置哪些可以修改标题
                if (d->editableContainerId.contains(launcherItem->id())) {
                    launcherItem->setEditable(true);
                }

                d->groupPageMaxItemCount[launcherItem->id()] =
                        (launcherItem->rowSpan() - 1) * launcherItem->columnSpan();
                d->groupPageMaxItemCount.insert(containerItem->getId(),
                                            (launcherItem->rowSpan()-1) * launcherItem->columnSpan());

                MultiPageModel *groupModel = new MultiPageModel;
                QList<QList<quint32>> childItemsId = setData.value(containerItem->getId());
                for (int page=0; page<childItemsId.count(); ++page) {
                    PageModel *groupPageModel = new PageModel;
                    int itemCount = childItemsId.at(page).count();
                    for (int itemIndex=0; itemIndex<itemCount; ++itemIndex) {
                        LauncherItem *groupItem = new LauncherItem;
                        UKUI::IconItem *appItem =
                                static_cast<UKUI::IconItem *>(itemData[childItemsId.at(page).at(itemIndex)]);
                        if (appItem) {
                            groupItem->setId(appItem->getId());
                            groupItem->setName(appItem->getName());
                            groupItem->setType(LauncherItem::AppIcon);
                            groupItem->setPageIndex(page);
                            groupItem->setIcon(appItem->getIconName());
                            groupItem->setDesktopFile(appItem->getDesktopName());
                            groupItem->setRowSpan(1);
                            groupItem->setColumnSpan(1);
                            groupPageModel->push_back(groupItem);
                        }
                    }
                    if (d->needAddIcon &&
                            groupPageModel->count() < d->groupPageMaxItemCount[containerItem->getId()] &&
                            d->editableContainerId.contains(containerItem->getId())) {
                        LauncherItem *addItem = d->createAddIconItem(page);
                        groupPageModel->push_back(addItem);
                    }
                    groupModel->push_back(groupPageModel);
                }
                d->groupdModel.insert(containerItem->getId(), groupModel);
                break;
            }

            case UKUI::Type::Widget: {
                UKUI::Widget *widgetItem = static_cast<UKUI::Widget *>(item);
                if (!widgetItem) {
                    break;
                }
                launcherItem->setId(widgetItem->getId());
                launcherItem->setName(widgetItem->getName());
                launcherItem->setType(LauncherItem::Widget);
                launcherItem->setPageIndex(i.key() - 1);
                launcherItem->setRowSpan(widgetItem->getPlaceHolder().at(0));
                launcherItem->setColumnSpan(widgetItem->getPlaceHolder().at(1));
                break;
            }

            case UKUI::Type::Unknown:
                break;
            default:
                break;
            }
            pageModel->push_back(launcherItem);
        }
        d->launcherModel.push_back(pageModel);
        ++i;
    }
    // init task bar
    QList<WindowId> runningAppWindows = WindowManager::windows();
    for (auto &wid : runningAppWindows) {
        QString desktopName = d->appManager->AppDesktopFileNameByPid(WindowManager::getPid(wid));
//        qDebug() << "init taskbar openwindow wid,pid--->" << wid << "--->" << WindowManager::getPid(wid) <<"desktopname-->" << desktopName;
        if(desktopName.isEmpty()) {
            continue;
        }
        QString desktopname = findDesktopNameFromTaskbar(desktopName);
        if (desktopname.isEmpty()) {
            qDebug() << "desktop cannot be found from taskbar";
            desktopname = findDesktopName(desktopName);
            if(!desktopname.isEmpty()) {
                QString filePathName = findIconPath(desktopname);
                d->taskBarManager->setWindowDesktopFileName(wid.toInt(), desktopname);
                addWindowIconToTaskbar(filePathName, desktopname, wid.toInt());
            }
        } else {
            //该应用在任务栏，设置为打开状态
            setAppOpenState(desktopname, wid.toInt());
            d->taskBarManager->setWindowDesktopFileName(wid.toInt(), desktopname);
        }
    }
}

void ModelManager::categories(const QMap<UKUI::Categories, QList<quint32>> categories)
{
    Q_D(ModelManager);

    d->categoryModel.clear();
    d->categoryItemsModel.clear();

    CategoryItem *allCategoryItem = new CategoryItem;
    int pageCount = d->dataHandler->apps().count() % d->pageItemMaxOfAddition ?
                d->dataHandler->apps().count() / d->pageItemMaxOfAddition + 1 :
                d->dataHandler->apps().count() / d->pageItemMaxOfAddition;
    allCategoryItem->setCategory(CategoryItem::All);
    allCategoryItem->setName("全部");
    allCategoryItem->setNormalIcon(d->categoryIcon(CategoryItem::All).first);
    allCategoryItem->setClickIcon(d->categoryIcon(CategoryItem::All).second);
    allCategoryItem->setPageCount(pageCount);
    d->categoryModel.push_back(allCategoryItem);

    // All
    auto apps = d->dataHandler->apps();
    auto appIt = apps.constBegin();
    PageModel *pageModel = new PageModel;
    while (appIt != apps.constEnd()) {
        UKUI::IconItem *iconItem =
                static_cast<UKUI::IconItem *>(d->dataHandler->items().value(*appIt));
        LauncherItem *launcherItem = new LauncherItem;
        launcherItem->setId(*appIt);
        launcherItem->setName(iconItem->getName());
        launcherItem->setIcon(iconItem->getIconName());
        launcherItem->setPageIndex(d->categoryItemsModel.isEmpty() ? 0 : d->categoryItemsModel.count());
        pageModel->push_back(launcherItem);
        if (pageModel->count() == d->pageItemMaxOfAddition) {
            d->categoryItemsModel.push_back(pageModel);
            pageModel = new PageModel;
        }
        ++ appIt;
    }
    if (!pageModel->isEmpty()) {
        d->categoryItemsModel.push_back(pageModel);
    } else {
        pageModel->deleteLater();
    }
    //

    auto it = categories.constBegin();
    while (it != categories.constEnd()) {
        if (it.value().isEmpty()) {
            ++ it;
            continue;
        }

        int category = static_cast<int>(it.key());
        CategoryItem *categoryItem = new CategoryItem;
        int pageCount = it.value().count() % d->pageItemMaxOfAddition ?
                    it.value().count() / d->pageItemMaxOfAddition + 1 :
                    it.value().count() / d->pageItemMaxOfAddition;
        categoryItem->setCategory(category + 1);
        categoryItem->setName(d->dataHandler->categoryName(static_cast<int>( it.key())));
        categoryItem->setPageCount(pageCount);
        categoryItem->setNormalIcon(d->categoryIcon(categoryItem->category()).first);
        categoryItem->setClickIcon(d->categoryIcon(categoryItem->category()).second);

        auto itemIt = it.value().constBegin();
        PageModel *pageModel = new PageModel;
        while (itemIt != it.value().constEnd()) {
            UKUI::IconItem *iconItem =
                    static_cast<UKUI::IconItem *>(d->dataHandler->items().value(*itemIt));
            LauncherItem *launcherItem = new LauncherItem;
            launcherItem->setName(iconItem->getName());
            launcherItem->setIcon(iconItem->getIconName());
            launcherItem->setId(*itemIt);
            launcherItem->setPageIndex(d->categoryItemsModel.isEmpty() ? 0 : d->categoryItemsModel.count());
            pageModel->push_back(launcherItem);
            if (pageModel->count() == d->pageItemMaxOfAddition) {
                d->categoryItemsModel.push_back(pageModel);
                pageModel = new PageModel;
            }
            ++ itemIt;
        }
        if (!pageModel->isEmpty()) {
            d->categoryItemsModel.push_back(pageModel);
        } else {
            pageModel->deleteLater();
        }

        d->categoryModel.push_back(categoryItem);
        ++ it;
    }
}

void ModelManager::removeItemFromGroup(int itemId, int groupId)
{
    Q_D(ModelManager);
    d->removeGroupItem(itemId, groupId);
    d->removeItemFromCategory(itemId);
}

void ModelManager::removeItemWithoutGroup(int itemId)
{
    Q_D(ModelManager);
    d->removeItemFromCategory(itemId);
    if (!d->needAddIcon) {
        for (int page=0; page<d->launcherModel.count(); ++page) {
            for (int itemIndex=0; itemIndex<d->launcherModel.at(page)->count(); ++itemIndex) {
                if (d->launcherModel.at(page)->at(itemIndex)->id() == itemId) {
                    removeItem(itemIndex, page);
                    return;
                }
            }
        }
    }
}

void ModelManager::itemInstalled(int itemId,
                                 int indexOfAllCategory,
                                 int indexOfCategory,
                                 int page,
                                 bool newCategory,
                                 int category)
{
    Q_D(ModelManager);

    UKUI::IconItem *iconItem =
            static_cast<UKUI::IconItem *>(d->dataHandler->items().value(itemId));
    LauncherItem *launcherItem = new LauncherItem;
    launcherItem->setName(iconItem->getName());
    launcherItem->setIcon(iconItem->getIconName());
    launcherItem->setId(itemId);
    launcherItem->setRowSpan(1);
    launcherItem->setColumnSpan(1);
    launcherItem->setPageIndex(page);
    launcherItem->setDesktopFile(iconItem->getDesktopName());

    // xc通用平板
    if (!d->needAddIcon && page != -1) {
        if (d->launcherModel.count() > page) {
            d->launcherModel.at(page)->push_back(launcherItem);
        } else if (d->launcherModel.count() == page) {
            PageModel *newPage = new PageModel;
            newPage->push_back(launcherItem);
            d->launcherModel.push_back(newPage);
        }
        return;
    }
    if (CategoryItem::All == d->currentCategoryOfAddion) {
        launcherItem->setState(LauncherItem::Selectable);
        launcherItem->setSelected(false);
    } else {
        launcherItem->setState(LauncherItem::Normal);
    }

    int insertAllCategoryPage = indexOfAllCategory / d->pageItemMaxOfAddition;
    int insertAllCategoryIndex = indexOfAllCategory % d->pageItemMaxOfAddition;        

    // new All category page
    if (d->categoryModel.first()->pageCount() <= insertAllCategoryPage) {
        d->categoryItemsModel.insert(insertAllCategoryPage, new PageModel);
        d->categoryItemsModel.at(insertAllCategoryPage)->push_back(launcherItem);
        launcherItem->setPageIndex(insertAllCategoryPage);
        int newPageCount = d->categoryModel.first()->pageCount() + 1;
        d->categoryModel.first()->setPageCount(newPageCount);
        d->updateCategoriesItemsPageIndex(insertAllCategoryPage + 1, 1);
        Q_EMIT categoryPageAdded(CategoryItem::All);
    } else {
        d->categoryItemsModel.at(insertAllCategoryPage)->insert(insertAllCategoryIndex, launcherItem);
        launcherItem->setPageIndex(insertAllCategoryPage);
        int page = insertAllCategoryPage;
        while (d->categoryItemsModel.at(page)->count() > d->pageItemMaxOfAddition &&
               (page + 1) < d->categoryModel.first()->pageCount()) {
            auto item = d->categoryItemsModel.at(page)->takeLast();
            item->setPageIndex(item->pageIndex() + 1);
            d->categoryItemsModel.at(++page)->push_front(item);
        }
        if (d->categoryItemsModel.at(page)->count() > d->pageItemMaxOfAddition) {
            auto item = d->categoryItemsModel.at(page)->takeLast();
            d->categoryItemsModel.insert(++page, new PageModel);
            int newPageCount = d->categoryModel.first()->pageCount() + 1;
            d->categoryModel.first()->setPageCount(newPageCount);
            item->setPageIndex(item->pageIndex() + 1);
            d->categoryItemsModel.at(page)->push_front(item);
            d->updateCategoriesItemsPageIndex(page + 1, 1);
            Q_EMIT categoryPageAdded(CategoryItem::All);
        }
    }

    LauncherItem *categoryLauncherItem = new LauncherItem;
    *categoryLauncherItem = *launcherItem;
    if (d->currentCategoryOfAddion == category) {
        categoryLauncherItem->setState(LauncherItem::Selectable);
        categoryLauncherItem->setSelected(false);
    } else {
        categoryLauncherItem->setState(LauncherItem::Normal);
    }
    // create a new category
    if (newCategory) {
        int newPageIndex = 0;
        bool foundNewCategoryPos = false;

        CategoryItem *categoryItem = new CategoryItem;
        categoryItem->setCategory(category);
        categoryItem->setName(d->dataHandler->categoryName(category - 1));
        categoryItem->setPageCount(1);
        categoryItem->setNormalIcon(d->categoryIcon(categoryItem->category()).first);
        categoryItem->setClickIcon(d->categoryIcon(categoryItem->category()).second);

        for (int categoryIndex=0; categoryIndex<d->categoryModel.count() - 1; ++categoryIndex) {
            newPageIndex += d->categoryModel.at(categoryIndex)->pageCount();
            if (category > d->categoryModel.at(categoryIndex)->category() &&
                    category < d->categoryModel.at(categoryIndex + 1)->category()) {
                d->categoryModel.insert(categoryIndex + 1, categoryItem);
                foundNewCategoryPos = true;
                break;
            }
        }
        // the last category
        if (!foundNewCategoryPos) {
            d->categoryModel.push_back(categoryItem);
            ++ newPageIndex;
        }
        PageModel *newPage = new PageModel;
        categoryLauncherItem->setPageIndex(newPageIndex);
        newPage->push_back(categoryLauncherItem);
        d->categoryItemsModel.insert(newPageIndex, newPage);      
        Q_EMIT categoryAdded(category);
        newPageIndex = foundNewCategoryPos ? newPageIndex + 1 : newPageIndex;
        d->updateCategoriesItemsPageIndex(newPageIndex, 1);
    } else {
        int insertCategoryPage = indexOfCategory / d->pageItemMaxOfAddition;
        int insertCategoryIndex = indexOfCategory % d->pageItemMaxOfAddition;
        int insertPage = insertCategoryPage;
        for (int index=0; index<d->categoryModel.count(); ++index) {
            if (d->categoryModel.at(index)->category() == category) {
                // insert to a new page
                if (d->categoryModel.at(index)->pageCount() <= insertCategoryPage) {
                    PageModel *pageModel = new PageModel;
                    pageModel->push_back(categoryLauncherItem);
                    if (d->categoryItemsModel.count() > insertPage) {
                        categoryLauncherItem->setPageIndex(insertPage);
                        d->categoryItemsModel.insert(insertPage, pageModel);
                        d->updateCategoriesItemsPageIndex(insertPage + 1, 1);
                    } else {
                        categoryLauncherItem->setPageIndex(d->categoryItemsModel.count());
                        d->categoryItemsModel.push_back(pageModel);
                    }
                    Q_EMIT categoryPageAdded(d->categoryModel.at(index)->category());
                }
                // insert to a already page
                else {
                    categoryLauncherItem->setPageIndex(insertPage);
                    d->categoryItemsModel.at(insertPage)->insert(insertCategoryIndex, categoryLauncherItem);
                    do {
                        if (d->categoryItemsModel.at(insertPage)->count() > d->pageItemMaxOfAddition) {
                            auto item = d->categoryItemsModel.at(insertPage)->takeLast();
                            item->setPageIndex(item->pageIndex() + 1);
                            // the last page of this category
                            if (d->categoryModel.at(index)->pageCount() == insertCategoryPage + 1) {
                                PageModel *pageModel = new PageModel;
                                pageModel->push_back(item);
                                d->categoryItemsModel.insert(insertPage + 1, pageModel);
                                d->updateCategoriesItemsPageIndex(insertPage + 2, 1);
                                d->categoryModel.at(index)->setPageCount(d->categoryModel.at(index)->pageCount() + 1);
                                Q_EMIT categoryPageAdded(d->categoryModel.at(index)->category());
                                break;
                            } else {
                                d->categoryItemsModel.at(insertPage ++)->push_front(item);
                            }
                        }
                        qDebug() << "do while";
                    } while(insertPage < d->categoryModel.at(index)->pageCount());
                }
                break;
            }
            insertPage += d->categoryModel.at(index)->pageCount();
        }
    }
}

void ModelManager::iconThemeChanged()
{
    Q_D(ModelManager);
    auto itemsData = d->dataHandler->items();
    auto updateIconFunction = [&itemsData](LauncherItem *item){
        UKUI::IconItem *iconItem =
                static_cast<UKUI::IconItem *>(itemsData.value(item->id()));
        if (iconItem) {
            item->setIcon(iconItem->getIconName());
        } else {
            qWarning() << "static_cast to UKUI::IconItem * failed: "
                       << item->id();
        }
    };
    auto groupIt = d->groupdModel.begin();
    while (groupIt != d->groupdModel.end()) {
        for (auto pageModel : *groupIt.value()) {
            for (auto item : *pageModel) {
                if (item->id() == 0) {
                    continue;
                }
                updateIconFunction(item);
            }
        }
        ++ groupIt;
    }

#ifdef EDU_SCREEN
    auto categoryIt = d->categoryItemsModel.begin();
    while (categoryIt != d->categoryItemsModel.end()) {
        for (auto item : (**categoryIt)) {
            if (item->id() == 0) {
                continue;
            }
            UKUI::IconItem *iconItem =
                    static_cast<UKUI::IconItem *>(itemsData.value(item->id()));
            if (iconItem) {
                item->setIcon(iconItem->getIconName());
            } else {
                qWarning() << "static_cast to category item UKUI::IconItem * failed: "
                           << item->id();
            }
        }
        ++ categoryIt;
    }
#else
    for (int page=0; page<d->launcherModel.count(); ++page) {
        for (int itemIndex=0; itemIndex<d->launcherModel.at(page)->count(); ++itemIndex) {
            if (d->launcherModel.at(page)->at(itemIndex)->type() == LauncherItem::AppIcon) {
                updateIconFunction(d->launcherModel.at(page)->at(itemIndex));
            }
        }
    }
#endif
}

void ModelManager::pageRemoved(int page, int groupId)
{
//    qDebug() << "remove page desktop" << page << groupId;
//    Q_D(ModelManager);
//    MultiPageModel *model = groupId >= 0 ? d->groupdModel.value(groupId) : &d->launcherModel;
//    if (model->at(page)->count() == 1) {
//        model->removeAt(page);
//        updateCachedData(-1);
//    }
}

void ModelManager::changeTaskbarData(appChangeInfo appChangeData)
{
    Q_D(ModelManager);

    //任务栏移除应用
    if (appChangeData.isTaskbarDeleteApp && !appChangeData.isDesktopDeleteApp && !appChangeData.isExchangePosition) {
        LauncherItem *launcherItem = d->groupdModel.value(0)->at(0)->takeAt(appChangeData.itemFromPageIndex);
        launcherItem->setPageIndex(appChangeData.itemToPage);

        if (d->launcherModel.count() == appChangeData.itemToPage) {
            PageModel *newPage = new PageModel;
            newPage->push_back(launcherItem);
            d->launcherModel.push_back(newPage);
            qDebug() << "添加新页--launchermodel.count-->" << d->launcherModel.count();
        } else {
            d->launcherModel.at(appChangeData.itemToPage)->insert(appChangeData.itemToPageIndex, launcherItem);
        }
//        updateCachedData(-1);
    }

    //任务栏添加应用
    if (!appChangeData.isTaskbarDeleteApp && appChangeData.isDesktopDeleteApp && !appChangeData.isExchangePosition) {
        LauncherItem *launcherItem = d->launcherModel.at(appChangeData.itemFromPage)->takeAt(appChangeData.itemFromPageIndex);
        if (d->launcherModel.at(appChangeData.itemFromPage)->count() == 0) {
            d->launcherModel.removeAt(appChangeData.itemFromPage);
            qDebug() << "删除空白页 pageid-->" << appChangeData.itemFromPage;
        }

        launcherItem->setPageIndex(0);
        d->groupdModel.value(0)->at(0)->insert(appChangeData.itemToPageIndex, launcherItem);
//        updateCachedData(-1);
    }

    //任务栏交换应用
    if (!appChangeData.isTaskbarDeleteApp && !appChangeData.isDesktopDeleteApp && appChangeData.isExchangePosition) {
        d->groupdModel.value(0)->at(0)->move(appChangeData.itemFromPageIndex, appChangeData.itemToPageIndex);
    }
}
