/*
 * Copyright 2022 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef MODELMANAGER_H
#define MODELMANAGER_H

#include <QObject>
#include "base/basemodel.h"
#include "base/categoryitem.h"
#include "dbusinterfaces/appmanagerinterface.h"

#include "types.h"
#include "kylin-tablet-desktop-backend.h"
#include "windowmanager/windowmanager.h"

using namespace kdk;

class LauncherItem;
class ModelManagerPrivate;

typedef BaseModel<LauncherItem *> PageModel;
typedef BaseModel<PageModel *> MultiPageModel;
typedef BaseModel<CategoryItem *> CategoryModel;
class ModelManager : public QObject
{
    Q_OBJECT
    Q_DISABLE_COPY(ModelManager)
    Q_DECLARE_PRIVATE(ModelManager)
    Q_PROPERTY(QAbstractListModel * launcherModel READ launcherModel NOTIFY launcherModelChanged)
    Q_PROPERTY(QAbstractListModel * categoryModel READ categoryModel NOTIFY categoryModelChanged)
    Q_PROPERTY(QAbstractListModel * categoryItemsModel READ categoryItemsModel NOTIFY categoryItemsModelChanged)
    Q_PROPERTY(QAbstractListModel * taskbarModel READ taskbarModel NOTIFY taskbarModelChanged)


public:
    explicit ModelManager(QObject *parent = nullptr);
    ~ModelManager();

    enum ItemParentType {
        Page,
        Group,
        Container,
    };

    ///
    /// \brief launcherModel 获取整个桌面的整体model
    /// \return
    ///
    Q_INVOKABLE QAbstractListModel* launcherModel() const;

    ///
    /// \brief groupModel 获取每个分组（或者Container）的model
    /// \param id 分组或者Container的id
    /// \return
    ///
    Q_INVOKABLE QAbstractListModel* groupModel(const quint32 &id) const;

    ///
    /// \brief categoryModel 返回应用分类的model
    /// \return
    ///
    Q_INVOKABLE QAbstractListModel* categoryModel() const;

    ///
    /// \brief categoryItemsModel 返回所有的分类的对应的item的model
    /// \return
    ///
    Q_INVOKABLE QAbstractListModel* categoryItemsModel() const;

    ///
    /// \brief groupPageItemCount 获取分组或者Container指定页的item的数量
    /// \param groupId 分组或者Container的id
    /// \param pageIndex 分组或者Container的页
    /// \return
    ///
    Q_INVOKABLE int groupPageItemCount(int groupId, int pageIndex);

    ///
    /// \brief launcerPageItemCount 返回当前页的item的数量
    /// \param pageIndex
    /// \return
    ///
    Q_INVOKABLE int launcerPageItemCount(int pageIndex) const;

    ///
    /// \brief moveItem 移动 item,\note 只适用于同一页面中的 item 移动.
    ///  跨页拖动时, 数据对于 GridView 来说是不同的 model, 因此并不适合 move.
    /// \param from 开始移动的 item 的 index
    /// \param to 要移动到的位置的 item 的 index
    /// \param page 当前移动的 item 所在的页的 index
    /// \param groupId  默认为 -1, 即不是分组内的 item
    /// \todo group->page的移动和page->group的移动
    ///
    Q_INVOKABLE void moveItem(int from, int to, int page, int groupId = -1);

    ///
    /// \brief addPlaceholderItem 添加占位 item, 跨页拖动 item 时（未释放）,
    /// 首先在目标 page 添加一个占位 item 来指示将要添加到的位置
    /// \param index 占位 item 所在页的 item index
    /// \param pageIndex 占位 item 所在的 page index
    /// \param rowSpan 占位 item 所跨的行数
    /// \param columnSpan 占位 item 所跨的列数
    /// \param groupId 当前所在的分组 id
    ///
    Q_INVOKABLE void addPlaceholderItem(int index,
                                        int pageIndex,
                                        int rowSpan = 1,
                                        int columnSpan = 1,
                                        int groupId = -1);

    ///
    /// \brief removePlaceholderItem 删除占位 item, 比如从第 0 页拖动 item 至第 1 页,
    /// 此时调用 \memberof addPlaceholderItem, 未释放 item 接着返回第 0 页, 则调用该函数删除占位 item
    ///
    Q_INVOKABLE void removePlaceholderItem();

    ///
    /// \brief replacePlaceholderItem 替换占位 item, 当移动 item 至相应位置释放该 item 时, 调用该函数,
    ///  将拖动的 item 的数据拷贝至占位 item
    /// \param fromIndex 拖动的 item 的 index
    /// \param fromPage 拖动的 item 所在的页
    /// \param groupId 拖动的 item 所在的分组, 默认为 -1, 即不是分组内的 item
    ///
    Q_INVOKABLE void replacePlaceholderItem(int fromIndex, int fromPage, int groupId = -1);

    ///
    /// \brief replacePlaceholderItemTaskbar 替换占位 item(占位 item与拖动 item的groupId不同), 当移动 item 至相应位置释放该 item 时, 调用该函数,
    ///  将拖动的 item 的数据拷贝至占位 item
    /// \param fromIndex 拖动的 item 的 index
    /// \param fromPage 拖动的 item 所在的页
    /// \param fromgroupId 拖动的 item 所在的分组, 默认为 -1, 即不是分组内的 item
    /// \param togroupId 占位 item 所在的分组, 默认为 -1
    ///
    Q_INVOKABLE void replacePlaceholderItemTaskbar(int fromIndex, int fromPage, int fromgroupId = -1, int togroupId= -1);

    ///
    /// \brief removeItem 删除 item, 比如卸载应用完成之后
    /// \param index item 的 index
    /// \param page item 所在页的 index
    /// \param groupId 如果 item 不在 Group 中, 则为 -1
    ///
    Q_INVOKABLE void removeItem(int index, int page, int groupId = -1);

    ///
    /// \brief setItemState 设置item的状态, \see LauncherItem::ItemSate
    /// \param state LauncherItem::ItemSate
    /// \param groupId
    ///
    Q_INVOKABLE void setItemState(int state, int groupId = -1);

    ///
    /// \brief setGroupName 设置分组的名称
    /// \param name 分组名称
    /// \param groupId 分组id
    ///
    Q_INVOKABLE void setGroupName(const QString &name, int groupId);

    ///
    /// \brief setAddAppCategory 设置添加应用界面的数据
    /// \param category 默认打卡的分类
    /// \param groupId 添加应用的分组
    ///
    Q_INVOKABLE void setAddAppCategory(int category, int groupId);

    ///
    /// \brief updateCachedData 每次操作完成之后，需要前端主动调用该函数进行缓存数据更新至数据库
    /// \param groupId 当前操作的分组id
    ///
    Q_INVOKABLE void updateCachedData(int groupId = -1);

    ///
    /// \brief groupEditable 分组是否可编辑，包括名称、删除和添加应用
    /// \param groupId 分组的id
    /// \return
    ///
    Q_INVOKABLE bool groupEditable(int groupId);

    ///
    /// \brief updateGroupItems 更新分组item的数据信息，主要用于添加应用界面的数据操作
    ///  每次点击“确认”按钮时需要调用该函数进行数据的更新
    /// \param groupId 当前分组的id
    /// \param pageIndex 打开添加应用界面时所在的page
    ///
    Q_INVOKABLE void updateGroupItems(int groupId, int pageIndex);

    Q_INVOKABLE void setItemSelectState(int itemId, int pageIndex, bool selected);

    ///获取任务栏model
    Q_INVOKABLE QAbstractListModel* taskbarModel() ;//const;

    ///
    /// \brief setAppOpenState 设置item为打开状态
    /// \param desktopName item的desktopName
    /// \param wid item的wid
    ///
    void setAppOpenState(QString desktopName, int wid);

    ///
    /// \brief setAppCloseState 设置item为关闭状态
    /// \param desktopName item的desktopName
    ///
    void setAppCloseState(QString desktopName);

    ///
    /// \brief addWindowIconToTaskbar 打开一个桌面应用，将桌面item图标加入任务栏
    /// \param iconPath item的图标路径
    /// \param desktopName item的desktopName
    /// \param wid item的wid
    ///
    void addWindowIconToTaskbar(QString iconPath, QString desktopName, int wid);


    void removeWindowIconFromTaskbar(QString iconPath);

    ///
    /// \brief removeAppIconFromTaskBar 关闭一个桌面应用，将桌面item图标从任务栏移除
    /// \param desktopFileName item的desktopFileName
    ///
    void removeAppIconFromTaskBar(const QString &desktopFileName);

    ///
    /// \brief findIconPath 根据desktopName查找item的图标路径
    /// \param desktopName item的desktopName
    ///
    QString findIconPath(QString desktopName);

    ///
    /// \brief findDesktopName 根据desktopName从桌面查找完整的desktop路径
    /// \param desktopName item的desktopName
    ///
    QString findDesktopName(QString desktopName);

    ///
    /// \brief findDesktopNameFromTaskbar 根据desktopName从任务栏查找完整的desktop路径
    /// \param desktopName item的desktopName
    ///
    QString findDesktopNameFromTaskbar(QString desktopName);

    ///
    /// \brief activeWindow 激活窗口
    /// \param wid item的wid
    ///
    Q_INVOKABLE void activeWindow(quint32 wid);

    ///
    /// \brief changeAppWindowWid 修改item保存的wid
    /// \param desktopName item被移除的wid
    /// \param newWid item保存的wid
    ///
    void changeAppWindowWid(QString desktopName, int wid, int newWid);

    void setPageModel(const MultiPageModel &pageModel);
    void setGroupModel(const QMap<int, MultiPageModel *> &groupModel);

Q_SIGNALS:
    void launcherModelChanged();
    void categoryModelChanged();
    void categoryItemsModelChanged();
    void categoryRemoved(int category);
    void categoryAdded(int category);
    void categoryPageAdded(int category);
    void categoryPageRemoved(int category);
    void taskbarModelChanged();
#if 0
    void itemMoved(int from, int to, int page, int groupId);
    void itemRemoved(int index, int page, int groupId);
    void itemInserted(int index, int page, int itemId, int groupId);
#endif
    void groupNameChanged(const QString &title, int groupId);

    void updateGroup(const QList<QList<quint32> > &ids, int groupId);

private Q_SLOTS:
    void desktopData(const QMap<quint32, QList<quint32>> &pageData,
                     const QMap<quint32, QList<QList<quint32>>> &setData,
                     const QMap<quint32, UKUI::BaseItem *> &itemData);

    void categories(const QMap<UKUI::Categories, QList<quint32>> categories);

    void removeItemFromGroup(int itemId, int groupId);
    void removeItemWithoutGroup(int itemId);
    void itemInstalled(int itemId,
                       int indexOfAllCategory,
                       int indexOfCategory,
                       int page = -1,
                       bool newCategory = false,
                       int category = CategoryItem::All);
    void iconThemeChanged();
    void pageRemoved(int page, int groupId);
    void changeTaskbarData(appChangeInfo appChangeData);

private:
    ModelManagerPrivate *d_ptr;
    AppManagerInterface *appManagerInterface;

};

#endif // MODELMANAGER_H
